# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the lintian workflow."""

import datetime as dt
from collections.abc import Sequence
from typing import Any, ClassVar, override
from unittest import mock

from django.test import override_settings
from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianLintian,
    DebianLintianSummary,
    TaskTypes,
)
from debusine.client.models import LookupChildType
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    Collection,
    CollectionItem,
    WorkRequest,
)
from debusine.db.models.work_requests import SkipWorkRequest
from debusine.server.workflows import LintianWorkflow, WorkflowValidationError
from debusine.server.workflows.models import (
    BaseWorkflowData,
    LintianWorkflowData,
    LintianWorkflowDynamicData,
    RegressionTrackingUpdateQAResults,
    SbuildWorkflowData,
)
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestCase,
)
from debusine.server.workflows.tests.test_regression_tracking import (
    RegressionTrackingWorkflowTestMixin,
)
from debusine.tasks.lintian import Lintian
from debusine.tasks.models import (
    ActionSkipIfLookupResultChanged,
    ActionUpdateCollectionWithArtifacts,
    BackendType,
    BaseDynamicTaskData,
    InputArtifactMultiple,
    InputArtifactSingle,
    LintianFailOnSeverity,
    LintianOutput,
    LookupMultiple,
    OutputData,
    RegressionAnalysisStatus,
    SbuildData,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry


class LintianWorkflowTests(
    RegressionTrackingWorkflowTestMixin, WorkflowTestCase[LintianWorkflow]
):
    """Unit tests for :py:class:`LintianWorkflow`."""

    SAMPLE_TASK_DATA = {
        "vendor": "debian",
        "codename": "trixie",
    }

    source_artifact: ClassVar[Artifact]
    binary_artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        cls.binary_artifact = (
            cls.playground.create_minimal_binary_package_artifact()
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data["source_artifact"] = self.source_artifact.id
        self.task_data["binary_artifacts"] = LookupMultiple(
            (self.binary_artifact.id,)
        )

    def create_lintian_workflow(
        self,
        *,
        extra_task_data: dict[str, Any] | None = None,
        validate: bool = True,
    ) -> LintianWorkflow:
        """Create a lintian workflow."""
        binary_artifact_1 = (
            self.playground.create_minimal_binary_package_artifact()
        )
        binary_artifact_2 = (
            self.playground.create_minimal_binary_package_artifact()
        )
        task_data = {
            "source_artifact": self.source_artifact.pk,
            "binary_artifacts": [
                "internal@collections/name:build-amd64",
                "internal@collections/name:build-i386",
            ],
            "vendor": "debian",
            "codename": "bookworm",
        }
        if extra_task_data is not None:
            task_data.update(extra_task_data)

        wr = self.playground.create_workflow(
            task_name="lintian",
            task_data=task_data,
            validate=validate,
            internal_collection_artifacts=[
                (binary_artifact_1, {"name": "build-amd64"}),
                (binary_artifact_2, {"name": "build-i386"}),
            ],
        )

        return self.get_workflow(wr)

    def _expected_task_display_name(
        self, architecture: str, lintian: WorkRequest
    ) -> str:
        """Return the expected display name for a lintian task."""
        display_name = f"Lintian for {architecture}"
        output = lintian.task_data["output"]

        binary_all_analysis = output.get("binary_all_analysis", True)
        source_analysis = output.get("source_analysis", True)

        if architecture != "all" and binary_all_analysis:
            display_name += "+all"

        if source_analysis:
            display_name += "+source"

        return display_name

    def orchestrate(
        self,
        task_data: LintianWorkflowData,
        architectures: Sequence[str],
        parent: WorkRequest | None = None,
        pipeline_task_name: str = "examplepipeline",
    ) -> WorkRequest:
        """Create and orchestrate a LintianWorkflow."""

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Lintian workflow."""

            TASK_NAME = pipeline_task_name

            @override
            def populate(self_) -> None:
                sbuild = self_.work_request.create_child_workflow(
                    task_name="sbuild",
                    task_data=SbuildWorkflowData(
                        input=SbuildInput(
                            source_artifact=task_data.source_artifact
                        ),
                        target_distribution="debian:sid",
                        architectures=list(architectures) or ["all"],
                    ),
                )
                self.playground.advance_work_request(sbuild, mark_running=True)

                for arch in architectures:
                    child = sbuild.create_child_worker(
                        task_name="sbuild",
                        task_data=SbuildData(
                            input=SbuildInput(
                                source_artifact=task_data.source_artifact
                            ),
                            build_architecture=arch,
                            environment="debian/match:codename=sid",
                        ),
                    )

                    self_.provides_artifact(
                        child,
                        ArtifactCategory.BINARY_PACKAGE,
                        f"build-{arch}",
                        data={"architecture": arch},
                    )
                sbuild.unblock_workflow_children()

                lintian = self_.work_request.create_child_workflow(
                    task_name="lintian",
                    task_data=task_data,
                )
                self.playground.advance_work_request(lintian, mark_pending=True)
                self_.orchestrate_child(lintian)

        root = self.playground.create_workflow(
            task_name=pipeline_task_name, parent=parent
        )
        self.assertTrue(
            self.playground.orchestrate_workflow(root, use_data=False)
        )

        return root

    def add_lintian_binary(
        self, suite: Collection, version: str, architecture: str = "all"
    ) -> None:
        """Add a lintian binary package to a suite."""
        artifact = self.playground.create_minimal_binary_package_artifact(
            "lintian", version, "lintian", version, architecture
        )
        suite.manager.add_artifact(
            artifact,
            user=self.playground.get_default_user(),
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )

    def add_qa_result(
        self,
        qa_results: Collection,
        package: str,
        version: str,
        architecture: str,
        lintian_version: str,
    ) -> CollectionItem:
        """Add a lintian result to a ``debian:qa-results`` collection."""
        work_request = self.playground.create_worker_task(
            task_name="lintian",
            result=WorkRequest.Results.SUCCESS,
            validate=False,
            use_data=False,
        )
        return qa_results.manager.add_artifact(
            self.playground.create_artifact(
                category=ArtifactCategory.LINTIAN,
                data=DebianLintian(
                    architecture=architecture,
                    summary=DebianLintianSummary(
                        tags_count_by_severity={},
                        package_filename={},
                        tags_found=[],
                        overridden_tags_found=[],
                        lintian_version=lintian_version,
                        distribution="debian:sid",
                    ),
                ),
            )[0],
            user=self.playground.get_default_user(),
            variables={
                "package": package,
                "version": version,
                "architecture": architecture,
                "timestamp": int(timezone.now().timestamp()),
                "work_request_id": work_request.id,
            },
        )

    def test_validate_input(self) -> None:
        """validate_input passes a valid case."""
        w = self.create_lintian_workflow()

        w.validate_input()

    def test_validate_input_bad_qa_suite(self) -> None:
        """validate_input raises errors in looking up a suite."""
        w = self.create_lintian_workflow(
            extra_task_data={"qa_suite": "nonexistent@debian:suite"},
            validate=False,
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:suite' does not exist or is hidden",
        ):
            w.validate_input()

    def test_validate_input_bad_reference_qa_results(self) -> None:
        """validate_input raises errors in looking up reference QA results."""
        w = self.create_lintian_workflow(
            extra_task_data={
                "reference_qa_results": "nonexistent@debian:qa-results"
            },
            validate=False,
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:qa-results' does not exist or is hidden",
        ):
            w.validate_input()

    def test_has_current_reference_qa_result_no_match(self) -> None:
        """_has_current_reference_qa_result: no matching result."""
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_lintian_binary(sid_suite, "2.116.3")
        self.add_qa_result(sid_qa_results, "other", "1.0-1", "amd64", "2.116.3")

        wr = self.playground.create_workflow(
            task_name="lintian",
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple(()),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result(
                architecture="amd64", build_architecture="amd64"
            )
        )

    def test_has_current_reference_qa_result_different_version(self) -> None:
        """_has_current_reference_qa_result: result for different version."""
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_lintian_binary(sid_suite, "2.116.3")
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64", "2.116.3")
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-2"
        )

        wr = self.playground.create_workflow(
            task_name="lintian",
            task_data=LintianWorkflowData(
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple(()),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result(
                architecture="amd64", build_architecture="amd64"
            )
        )

    def test_has_current_reference_qa_result_old_lintian_version(self) -> None:
        """_has_current_reference_qa_result: run with old lintian version."""
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_lintian_binary(sid_suite, "2.116.3")
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64", "2.15.0")

        wr = self.playground.create_workflow(
            task_name="lintian",
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple(()),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result(
                architecture="amd64", build_architecture="amd64"
            )
        )

    def test_has_current_reference_qa_result_no_lintian_version(self) -> None:
        """_has_current_reference_qa_result: no reference lintian version."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64", "2.116.3")

        wr = self.playground.create_workflow(
            task_name="lintian",
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple(()),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.assertTrue(
            self.get_workflow(wr)._has_current_reference_qa_result(
                architecture="amd64", build_architecture="amd64"
            )
        )

    def test_has_current_reference_qa_result_current(self) -> None:
        """_has_current_reference_qa_result: current result."""
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_lintian_binary(sid_suite, "2.116.3")
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64", "2.116.3")

        wr = self.playground.create_workflow(
            task_name="lintian",
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple(()),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.assertTrue(
            self.get_workflow(wr)._has_current_reference_qa_result(
                architecture="amd64", build_architecture="amd64"
            )
        )

    def test_has_current_reference_qa_result_honours_build_architecture(
        self,
    ) -> None:
        """_has_current_reference_qa_result: lintian checked by build arch."""
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_lintian_binary(sid_suite, "2.116.3")
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "all", "2.116.3")

        wr = self.playground.create_workflow(
            task_name="lintian",
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple(()),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.assertTrue(
            self.get_workflow(wr)._has_current_reference_qa_result(
                architecture="all", build_architecture="amd64"
            )
        )

        self.add_lintian_binary(sid_suite, "2.117.0", architecture="amd64")

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result(
                architecture="all", build_architecture="amd64"
            )
        )

    @preserve_db_task_registry()
    def test_populate(self) -> None:
        """Test populate."""
        architectures = ["amd64", "i386", "all"]

        root = self.orchestrate(
            task_data=LintianWorkflowData(
                architectures=architectures,
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                    ]
                ),
                vendor="debian",
                codename="trixie",
            ),
            architectures=architectures,
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        lintians = workflow.children.filter(task_name="lintian")

        # architecture "all" does not have a lintian job
        self.assertEqual(lintians.count(), len(architectures) - 1)

        for architecture in architectures:
            if architecture == "all":
                continue

            lintian = lintians.get(
                workflow_data_json__step=f"lintian-{architecture}"
            )

            self.assertEqual(
                lintian.task_data,
                {
                    "backend": Lintian.DEFAULT_BACKEND,
                    "build_architecture": architecture,
                    "environment": "debian/match:codename=trixie",
                    "exclude_tags": [],
                    "fail_on_severity": LintianFailOnSeverity.ERROR,
                    "include_tags": [],
                    "input": {
                        "binary_artifacts": sorted(
                            [
                                f"internal@collections/name:build-{arch}"
                                for arch in (architecture, "all")
                            ]
                        ),
                        "source_artifact": self.source_artifact.id,
                    },
                    "output": (
                        {"source_analysis": False, "binary_all_analysis": False}
                        if architecture != "amd64"
                        else {}
                    ),
                    "target_distribution": "debian:trixie",
                },
            )

            self.assertEqual(
                lintian.workflow_data_json,
                {
                    "display_name": self._expected_task_display_name(
                        architecture, lintian
                    ),
                    "step": f"lintian-{architecture}",
                },
            )

            self.assertQuerySetEqual(
                lintian.dependencies.all(),
                set(
                    WorkRequest.objects.filter(
                        task_name="sbuild",
                        task_data__build_architecture__in=(architecture, "all"),
                    )
                ),
                ordered=False,
            )

    @preserve_db_task_registry()
    def test_populate_source_only(self) -> None:
        """Test populate with no binary artifacts."""
        root = self.orchestrate(
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple(()),
                vendor="debian",
                codename="trixie",
            ),
            architectures=("all",),
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        lintian = workflow.children.get(task_name="lintian")

        self.assertEqual(
            lintian.task_data,
            {
                "backend": BackendType.UNSHARE,
                "build_architecture": "amd64",
                "environment": "debian/match:codename=trixie",
                "exclude_tags": [],
                "fail_on_severity": LintianFailOnSeverity.ERROR,
                "include_tags": [],
                "input": {
                    "binary_artifacts": [],
                    "source_artifact": self.source_artifact.id,
                },
                "output": {},
                "target_distribution": "debian:trixie",
            },
        )

    @preserve_db_task_registry()
    def test_populate_without_architectures(self) -> None:
        """Test populate with only arch-indep packages."""
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                "hello", "1.0.0", "hello", "1.0.0", "all"
            )
        )

        root = self.orchestrate(
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="trixie",
            ),
            architectures=("all",),
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        lintian = workflow.children.get(task_name="lintian")

        self.assertEqual(
            lintian.task_data,
            {
                "backend": BackendType.UNSHARE,
                "build_architecture": "amd64",
                "environment": "debian/match:codename=trixie",
                "exclude_tags": [],
                "fail_on_severity": LintianFailOnSeverity.ERROR,
                "include_tags": [],
                "input": {
                    "binary_artifacts": [binary_artifact.id],
                    "source_artifact": self.source_artifact.id,
                },
                "output": {},
                "target_distribution": "debian:trixie",
            },
        )

    @preserve_db_task_registry()
    def test_populate_without_architectures_arch_all_build_architecture(
        self,
    ) -> None:
        """Workflow honours `arch_all_build_architecture`."""
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                "hello", "1.0.0", "hello", "1.0.0", "all"
            )
        )

        root = self.orchestrate(
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="trixie",
                arch_all_build_architecture="s390x",
            ),
            architectures=("all",),
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        lintian = workflow.children.get(task_name="lintian")

        self.assertEqual(
            lintian.task_data,
            {
                "backend": BackendType.UNSHARE,
                "build_architecture": "s390x",
                "environment": "debian/match:codename=trixie",
                "exclude_tags": [],
                "fail_on_severity": LintianFailOnSeverity.ERROR,
                "include_tags": [],
                "input": {
                    "binary_artifacts": [binary_artifact.id],
                    "source_artifact": self.source_artifact.id,
                },
                "output": {},
                "target_distribution": "debian:trixie",
            },
        )

    @preserve_db_task_registry()
    def test_populate_mixed_without_arch_all_build_architecture(self) -> None:
        """Mixed all/any binaries but without the usual arch-all arch."""
        binary_artifacts = {
            "all": self.playground.create_minimal_binary_package_artifact(
                "hello", "1.0.0", "hello", "1.0.0", "all"
            ),
            "i386": self.playground.create_minimal_binary_package_artifact(
                "hello", "1.0.0", "libhello1", "1.0.0", "i386"
            ),
        }

        root = self.orchestrate(
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple(
                    (binary_artifacts["all"].id, binary_artifacts["i386"].id)
                ),
                vendor="debian",
                codename="trixie",
                arch_all_build_architecture="amd64",
            ),
            architectures=("all", "i386"),
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        self.assertQuerySetEqual(
            workflow.children.filter(task_name="lintian").values_list(
                "task_data", flat=True
            ),
            [
                {
                    "backend": BackendType.UNSHARE,
                    "build_architecture": "i386",
                    "environment": "debian/match:codename=trixie",
                    "exclude_tags": [],
                    "fail_on_severity": LintianFailOnSeverity.ERROR,
                    "include_tags": [],
                    "input": {
                        "binary_artifacts": sorted(
                            [
                                binary_artifacts["all"].id,
                                binary_artifacts["i386"].id,
                            ]
                        ),
                        "source_artifact": self.source_artifact.id,
                    },
                    "output": {},
                    "target_distribution": "debian:trixie",
                }
            ],
        )

    @preserve_db_task_registry()
    def test_populate_arch_dep_without_arch_all_build_architecture(
        self,
    ) -> None:
        """``Architecture: any`` binaries without the usual arch-all arch."""
        binary_artifacts = {
            "i386": self.playground.create_minimal_binary_package_artifact(
                "hello", "1.0.0", "libhello1", "1.0.0", "i386"
            ),
            "s390x": self.playground.create_minimal_binary_package_artifact(
                "hello", "1.0.0", "libhello1", "1.0.0", "s390x"
            ),
        }

        root = self.orchestrate(
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple(
                    (binary_artifacts["i386"].id, binary_artifacts["s390x"].id)
                ),
                vendor="debian",
                codename="trixie",
                arch_all_build_architecture="amd64",
            ),
            architectures=("i386", "s390x"),
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        self.assertQuerySetEqual(
            workflow.children.filter(task_name="lintian")
            .order_by("id")
            .values_list("task_data", flat=True),
            [
                {
                    "backend": BackendType.UNSHARE,
                    "build_architecture": architecture,
                    "environment": "debian/match:codename=trixie",
                    "exclude_tags": [],
                    "fail_on_severity": LintianFailOnSeverity.ERROR,
                    "include_tags": [],
                    "input": {
                        "binary_artifacts": [binary_artifacts[architecture].id],
                        "source_artifact": self.source_artifact.id,
                    },
                    "output": (
                        {"source_analysis": False, "binary_all_analysis": False}
                        if architecture != "i386"
                        else {}
                    ),
                    "target_distribution": "debian:trixie",
                }
                for architecture in ("i386", "s390x")
            ],
        )

    @preserve_db_task_registry()
    def test_populate_with_no_architecture_overlap(self) -> None:
        """If architectures and binary artifacts are disjoint, do nothing."""
        root = self.orchestrate(
            task_data=LintianWorkflowData(
                architectures=["i386"],
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple(
                    ("internal@collections/name:build-amd64",)
                ),
                vendor="debian",
                codename="trixie",
            ),
            architectures=["amd64"],
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        self.assertFalse(workflow.children.filter(task_name="lintian").exists())

    @preserve_db_task_registry()
    def test_populate_no_binary_any_analysis(self) -> None:
        """Disabling binary-any analysis skips no-op work requests."""
        binary_artifacts = {
            "all": self.playground.create_minimal_binary_package_artifact(
                "hello", "1.0.0", "hello", "1.0.0", "all"
            ),
            "amd64": self.playground.create_minimal_binary_package_artifact(
                "hello", "1.0.0", "libhello1", "1.0.0", "amd64"
            ),
            "i386": self.playground.create_minimal_binary_package_artifact(
                "hello", "1.0.0", "libhello1", "1.0.0", "i386"
            ),
        }

        root = self.orchestrate(
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple(
                    (
                        binary_artifacts["all"].id,
                        binary_artifacts["amd64"].id,
                        binary_artifacts["i386"].id,
                    )
                ),
                vendor="debian",
                codename="trixie",
                arch_all_build_architecture="amd64",
                output=LintianOutput(binary_any_analysis=False),
            ),
            architectures=("all", "amd64", "i386"),
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        self.assertQuerySetEqual(
            workflow.children.filter(task_name="lintian").values_list(
                "task_data", flat=True
            ),
            [
                {
                    "backend": BackendType.UNSHARE,
                    "build_architecture": "amd64",
                    "environment": "debian/match:codename=trixie",
                    "exclude_tags": [],
                    "fail_on_severity": LintianFailOnSeverity.ERROR,
                    "include_tags": [],
                    "input": {
                        "binary_artifacts": sorted(
                            [
                                binary_artifacts["all"].id,
                                binary_artifacts["amd64"].id,
                            ]
                        ),
                        "source_artifact": self.source_artifact.id,
                    },
                    "output": {"binary_any_analysis": False},
                    "target_distribution": "debian:trixie",
                }
            ],
        )

    @preserve_db_task_registry()
    def test_populate_upload(self) -> None:
        """The workflow accepts debian:upload source artifacts."""
        architectures = ["amd64", "i386", "all"]
        upload_artifacts = self.playground.create_upload_artifacts()

        root = self.orchestrate(
            task_data=LintianWorkflowData(
                architectures=architectures,
                source_artifact=upload_artifacts.upload.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                    ]
                ),
                vendor="debian",
                codename="trixie",
            ),
            architectures=architectures,
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        lintians = workflow.children.filter(task_name="lintian")

        # architecture "all" does not have a lintian job
        self.assertEqual(lintians.count(), len(architectures) - 1)

        for architecture in architectures:
            if architecture == "all":
                continue

            lintian = lintians.get(
                workflow_data_json__step=f"lintian-{architecture}"
            )

            self.assertEqual(
                lintian.task_data,
                {
                    "backend": Lintian.DEFAULT_BACKEND,
                    "build_architecture": architecture,
                    "environment": "debian/match:codename=trixie",
                    "exclude_tags": [],
                    "fail_on_severity": LintianFailOnSeverity.ERROR,
                    "include_tags": [],
                    "input": {
                        "binary_artifacts": sorted(
                            [
                                f"internal@collections/name:build-{arch}"
                                for arch in (architecture, "all")
                            ]
                        ),
                        "source_artifact": (
                            f"{upload_artifacts.source.id}@artifacts"
                        ),
                    },
                    "output": (
                        {"source_analysis": False, "binary_all_analysis": False}
                        if architecture != "amd64"
                        else {}
                    ),
                    "target_distribution": "debian:trixie",
                },
            )

            self.assertEqual(
                lintian.workflow_data_json,
                {
                    "display_name": self._expected_task_display_name(
                        architecture, lintian
                    ),
                    "step": f"lintian-{architecture}",
                },
            )

            self.assertQuerySetEqual(
                lintian.dependencies.all(),
                set(
                    WorkRequest.objects.filter(
                        task_name="sbuild",
                        task_data__build_architecture__in=(architecture, "all"),
                    )
                ),
                ordered=False,
            )

    def test_populate_uploads(self) -> None:
        """The workflow accepts debian:upload artifacts in all inputs."""
        artifacts = self.playground.create_upload_artifacts(
            binaries=[
                ("hello", "amd64"),
                ("hello", "i386"),
                ("hello-doc", "all"),
            ]
        )
        architectures = ["amd64", "i386"]

        wr = self.playground.create_workflow(
            task_name="lintian",
            task_data=LintianWorkflowData(
                source_artifact=artifacts.upload.id,
                binary_artifacts=LookupMultiple((artifacts.upload.id,)),
                architectures=architectures,
                vendor="debian",
                codename="sid",
            ),
        )
        self.get_workflow(wr).populate()
        children = list(wr.children.order_by("id").all())
        self.assertEqual(len(children), len(architectures))
        for architecture, child in zip(architectures, children):
            with self.subTest(architecture=architecture):
                self.assertEqual(
                    child.task_data["build_architecture"], architecture
                )
                self.assertEqual(
                    child.task_data["input"],
                    {
                        "source_artifact": (f"{artifacts.source.id}@artifacts"),
                        "binary_artifacts": sorted(
                            f"{artifact.id}@artifacts"
                            for artifact in artifacts.binaries
                            if artifact.data["deb_fields"]["Architecture"]
                            in (architecture, "all")
                        ),
                    },
                )

    @preserve_db_task_registry()
    def test_populate_has_current_reference_qa_result(self) -> None:
        """The workflow does nothing with a current reference QA result."""
        trixie_suite = self.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        trixie_qa_results = self.playground.create_collection(
            "trixie", CollectionCategory.QA_RESULTS
        )
        self.add_lintian_binary(trixie_suite, "2.116.3")

        architectures = ["amd64", "i386", "all"]
        for architecture in architectures:
            self.add_qa_result(
                trixie_qa_results,
                "hello",
                self.source_artifact.data["version"],
                architecture,
                "2.116.3",
            )

        root = self.orchestrate(
            task_data=LintianWorkflowData(
                prefix="reference-qa-result|",
                architectures=architectures,
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                    ]
                ),
                vendor="debian",
                codename="trixie",
                qa_suite=f"trixie@{CollectionCategory.SUITE}",
                reference_qa_results=f"trixie@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
            architectures=architectures,
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        self.assertQuerySetEqual(workflow.children.all(), [])

    @preserve_db_task_registry()
    def test_populate_has_current_reference_qa_result_force(self) -> None:
        """The workflow can be forced despite a current reference QA result."""
        trixie_suite = self.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        trixie_qa_results = self.playground.create_collection(
            "trixie", CollectionCategory.QA_RESULTS
        )
        self.add_lintian_binary(trixie_suite, "2.116.3")

        architectures = ["amd64", "i386", "all"]
        for architecture in architectures:
            self.add_qa_result(
                trixie_qa_results,
                "hello",
                self.source_artifact.data["version"],
                architecture,
                "2.116.3",
            )

        root = self.orchestrate(
            task_data=LintianWorkflowData(
                prefix="reference-qa-result|",
                architectures=architectures,
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                    ]
                ),
                vendor="debian",
                codename="trixie",
                qa_suite=f"trixie@{CollectionCategory.SUITE}",
                reference_qa_results=f"trixie@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.FORCE,
            ),
            architectures=architectures,
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        self.assertTrue(
            workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="lintian"
            ).exists()
        )

    @preserve_db_task_registry()
    def test_populate_no_previous_reference_qa_result(self) -> None:
        """The workflow produces a reference QA result if there is none."""
        trixie_suite = self.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        self.playground.create_collection(
            "trixie", CollectionCategory.QA_RESULTS
        )
        self.add_lintian_binary(trixie_suite, "2.116.3")

        architectures = ["amd64", "i386", "all"]

        source_item = trixie_suite.manager.add_artifact(
            self.source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )

        root = self.orchestrate(
            task_data=LintianWorkflowData(
                prefix="reference-qa-result|",
                architectures=architectures,
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                    ]
                ),
                vendor="debian",
                codename="trixie",
                qa_suite=f"trixie@{CollectionCategory.SUITE}",
                reference_qa_results=f"trixie@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
            architectures=architectures,
        )

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="lintian",
            parent=root,
        )

        lintians = workflow.children.filter(task_name="lintian")

        # architecture "all" does not have a lintian job
        concrete_architectures = [
            architecture
            for architecture in architectures
            if architecture != "all"
        ]
        self.assertEqual(lintians.count(), len(concrete_architectures))

        provide_actions: dict[str, ActionUpdateCollectionWithArtifacts] = {}
        skip_actions: dict[str, ActionSkipIfLookupResultChanged] = {}
        qa_result_actions: dict[str, ActionUpdateCollectionWithArtifacts] = {}
        results: dict[str, Artifact] = {}
        for architecture in concrete_architectures:
            lintian = lintians.get(
                workflow_data_json__step=f"lintian-{architecture}"
            )

            self.assertEqual(
                lintian.task_data,
                {
                    "backend": Lintian.DEFAULT_BACKEND,
                    "build_architecture": architecture,
                    "environment": "debian/match:codename=trixie",
                    "exclude_tags": [],
                    "fail_on_severity": LintianFailOnSeverity.ERROR,
                    "include_tags": [],
                    "input": {
                        "binary_artifacts": sorted(
                            [
                                f"internal@collections/name:build-{arch}"
                                for arch in (architecture, "all")
                            ]
                        ),
                        "source_artifact": self.source_artifact.id,
                    },
                    "output": (
                        {"source_analysis": False, "binary_all_analysis": False}
                        if architecture != "amd64"
                        else {}
                    ),
                    "target_distribution": "debian:trixie",
                },
            )

            self.assertEqual(
                lintian.workflow_data_json,
                {
                    "allow_failure": True,
                    "display_name": self._expected_task_display_name(
                        architecture, lintian
                    ),
                    "step": f"lintian-{architecture}",
                },
            )

            self.assertQuerySetEqual(
                lintian.dependencies.all(),
                set(
                    WorkRequest.objects.filter(
                        task_name="sbuild",
                        task_data__build_architecture__in=(architecture, "all"),
                    )
                ),
                ordered=False,
            )

            provide_actions[architecture] = ActionUpdateCollectionWithArtifacts(
                collection="internal@collections",
                name_template=f"reference-qa-result|lintian-{architecture}",
                artifact_filters={"category": ArtifactCategory.LINTIAN},
            )
            skip_actions[architecture] = ActionSkipIfLookupResultChanged(
                lookup=(
                    f"trixie@{CollectionCategory.QA_RESULTS}/"
                    f"latest:lintian_hello_{architecture}"
                ),
                collection_item_id=None,
                promise_name=f"reference-qa-result|lintian-{architecture}",
            )
            qa_result_actions[architecture] = (
                ActionUpdateCollectionWithArtifacts(
                    collection=f"trixie@{CollectionCategory.QA_RESULTS}",
                    variables={
                        "package": "hello",
                        "version": self.source_artifact.data["version"],
                        "$architecture": "architecture",
                        "timestamp": int(source_item.created_at.timestamp()),
                        "work_request_id": lintian.id,
                    },
                    artifact_filters={"category": ArtifactCategory.LINTIAN},
                )
            )
            self.assert_work_request_event_reactions(
                lintian,
                on_assignment=[skip_actions[architecture]],
                on_failure=[qa_result_actions[architecture]],
                on_success=[
                    provide_actions[architecture],
                    qa_result_actions[architecture],
                ],
            )

            # Completing the work request stores the QA result.
            artifact_architectures = (
                ["source", "all", architecture]
                if architecture == "amd64"
                else [architecture]
            )
            for artifact_architecture in artifact_architectures:
                results[artifact_architecture] = (
                    self.playground.create_artifact(
                        category=ArtifactCategory.LINTIAN,
                        data={"architecture": artifact_architecture},
                        work_request=lintian,
                    )[0]
                )
            self.playground.advance_work_request(
                lintian, result=WorkRequest.Results.SUCCESS, use_data=False
            )

        for architecture, result in results.items():
            self.assertEqual(
                lintian.lookup_single(
                    f"trixie@{CollectionCategory.QA_RESULTS}/"
                    f"latest:lintian_hello_{architecture}",
                    expect_type=LookupChildType.ARTIFACT,
                ).artifact,
                result,
            )

        # Population is idempotent, even if qa_suite changes.
        with (
            mock.patch.object(
                LintianWorkflow,
                "_has_current_reference_qa_result",
                return_value=False,
            ),
            mock.patch.object(
                LintianWorkflow,
                "qa_suite_changed",
                mock.PropertyMock(
                    return_value=timezone.now() + dt.timedelta(hours=1)
                ),
            ),
        ):
            self.get_workflow(workflow).populate()
        lintians = workflow.children.filter(task_name="lintian")
        self.assertEqual(lintians.count(), len(concrete_architectures))
        for architecture in concrete_architectures:
            lintian = lintians.get(
                workflow_data_json__step=f"lintian-{architecture}"
            )
            self.assert_work_request_event_reactions(
                lintian,
                on_assignment=[skip_actions[architecture]],
                on_failure=[qa_result_actions[architecture]],
                on_success=[
                    provide_actions[architecture],
                    qa_result_actions[architecture],
                ],
            )

    def test_populate_reference_qa_result_backs_off(self) -> None:
        """Reference tasks are skipped if another workflow got there first."""
        trixie_suite = self.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        trixie_qa_results = self.playground.create_collection(
            "trixie", CollectionCategory.QA_RESULTS
        )
        self.add_lintian_binary(trixie_suite, "2.116.3")
        self.playground.create_debian_environment(
            codename="trixie", variant="lintian"
        )

        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        wr = self.playground.create_workflow(
            task_name="lintian",
            task_data=LintianWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="trixie",
                qa_suite=f"trixie@{CollectionCategory.SUITE}",
                reference_qa_results=f"trixie@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.get_workflow(wr).populate()

        racing_qa_result = self.add_qa_result(
            trixie_qa_results, "hello", "1.0-1", "amd64", "2.116.3"
        )
        [child] = wr.children.all()
        self.playground.advance_work_request(child, mark_pending=True)

        with self.assertRaises(SkipWorkRequest):
            child.assign_worker(self.playground.create_worker())

        self.assertEqual(child.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(child.result, WorkRequest.Results.SKIPPED)
        self.assertEqual(
            child.output_data,
            OutputData(
                skip_reason=(
                    f"Result of lookup "
                    f"'trixie@{CollectionCategory.QA_RESULTS}/"
                    f"latest:lintian_hello_amd64' changed"
                )
            ),
        )
        self.assertEqual(
            child.lookup_single(
                f"trixie@{CollectionCategory.QA_RESULTS}/"
                f"latest:lintian_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).collection_item,
            racing_qa_result,
        )

    @preserve_db_task_registry()
    def test_orchestrate_idempotent(self) -> None:
        """Calling orchestrate twice does not create new work requests."""
        binary = self.playground.create_minimal_binary_package_artifact(
            "hello", "1.0.0", "hello", "1.0.0", "all"
        )

        wr = self.playground.create_workflow(
            task_name="lintian",
            task_data=LintianWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple((binary.id,)),
                vendor="debian",
                codename="trixie",
            ),
        )

        self.get_workflow(wr).populate()

        children = set(wr.children.all())

        self.get_workflow(wr).populate()

        self.assertQuerySetEqual(wr.children.all(), children, ordered=False)

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            LintianWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello_1.0-1",
                source_artifact_id=self.source_artifact.id,
                binary_artifacts_ids=[self.binary_artifact.id],
            ),
        )

    def test_get_input_artifacts(self) -> None:
        workflow = self.task
        self.assertIsNone(workflow.dynamic_data)

        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=None,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=None,
                ),
            ],
        )

        workflow.dynamic_data = LintianWorkflowDynamicData(
            source_artifact_id=self.source_artifact.id,
            binary_artifacts_ids=[self.binary_artifact.id],
        )

        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=self.source_artifact.id,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[self.binary_artifact.id],
                ),
            ],
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    @preserve_db_task_registry()
    def test_callback_regression_analysis(self) -> None:
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        self.playground.create_collection("sid", CollectionCategory.QA_RESULTS)
        architectures = ["amd64", "i386", "ppc64el", "s390x"]
        reference_source_artifact = self.playground.create_source_artifact(
            name="hello", version="0.9-1"
        )

        root = self.playground.create_workflow(mark_running=True)
        reference = self.orchestrate(
            task_data=LintianWorkflowData(
                prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                architectures=architectures,
                source_artifact=reference_source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in ("all", *architectures)
                        if arch != "ppc64el"
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            architectures=["all", *architectures],
            parent=root,
            pipeline_task_name="referencepipeline",
        )
        new = self.orchestrate(
            task_data=LintianWorkflowData(
                reference_prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                enable_regression_tracking=True,
                architectures=architectures,
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in ("all", *architectures)
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            architectures=["all", *architectures],
            parent=root,
        )

        # Unblock the lintian tasks.
        for sbuild in reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS, use_data=False
            )
        for sbuild in new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS, use_data=False
            )

        reference_lintian_workflow = reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="lintian"
        )
        reference_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in reference_lintian_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="lintian"
            )
        }
        reference_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.LINTIAN,
                data={
                    "architecture": architecture,
                    "summary": {
                        "tags_count_by_severity": tags_count_by_severity,
                        "tags_found": tags_found,
                    },
                },
                work_request=reference_tasks[
                    "amd64" if architecture == "all" else architecture
                ],
            )[0]
            for architecture, tags_count_by_severity, tags_found in (
                ("all", {"warning": 1, "error": 0}, ["warning-1"]),
                (
                    "amd64",
                    {"warning": 0, "error": 0},
                    [],
                ),
                (
                    "i386",
                    {"warning": 2, "error": 1},
                    ["warning-1", "warning-2", "error-1"],
                ),
                ("s390x", {"warning": 1, "error": 1}, ["warning-1", "error-1"]),
            )
        }
        for artifact in reference_results.values():
            self.playground.create_artifact_relation(
                artifact,
                reference_source_artifact,
                ArtifactRelation.Relations.RELATES_TO,
            )
        new_lintian_workflow = new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="lintian"
        )
        new_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in new_lintian_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="lintian"
            )
        }
        new_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.LINTIAN,
                data={
                    "architecture": architecture,
                    "summary": {
                        "tags_count_by_severity": tags_count_by_severity,
                        "tags_found": tags_found,
                    },
                },
                work_request=new_tasks[
                    "amd64" if architecture == "all" else architecture
                ],
            )[0]
            for architecture, tags_count_by_severity, tags_found in (
                ("all", {"warning": 1, "error": 0}, ["warning-1"]),
                (
                    "amd64",
                    {"warning": 0, "error": 0},
                    [],
                ),
                (
                    "i386",
                    {"warning": 1, "error": 2},
                    ["warning-1", "error-1", "error-2"],
                ),
                ("ppc64el", {"warning": 1, "error": 0}, ["warning-1"]),
                ("s390x", {"warning": 1, "error": 0}, ["warning-1"]),
            )
        }
        for artifact in new_results.values():
            self.playground.create_artifact_relation(
                artifact,
                self.source_artifact,
                ArtifactRelation.Relations.RELATES_TO,
            )

        self.assertEqual(
            new_lintian_workflow.output_data, OutputData(regression_analysis={})
        )
        self.schedule_and_run_workflow_callback(new_lintian_workflow)
        self.playground.advance_work_request(
            reference_tasks["amd64"],
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )
        self.schedule_and_run_workflow_callback(new_lintian_workflow)
        self.playground.advance_work_request(
            new_tasks["amd64"],
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )

        self.schedule_and_run_workflow_callback(new_lintian_workflow)

        expected_analysis: dict[
            str, tuple[RegressionAnalysisStatus, dict[str, list[str]]]
        ] = {
            "all": (
                RegressionAnalysisStatus.STABLE,
                {"added_tags": [], "removed_tags": []},
            ),
            "amd64": (
                RegressionAnalysisStatus.STABLE,
                {"added_tags": [], "removed_tags": []},
            ),
        }
        new_lintian_workflow.refresh_from_db()
        self.assert_regression_analysis(
            new_lintian_workflow,
            "0.9-1",
            reference_results,
            "1.0-1",
            new_results,
            expected_analysis,
        )

        self.assert_callback_status(
            new_lintian_workflow, WorkRequest.Statuses.BLOCKED
        )

        for architecture, reference_result, new_result, status, details in (
            (
                "i386",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.REGRESSION,
                {"added_tags": ["error-2"], "removed_tags": ["warning-2"]},
            ),
            (
                "ppc64el",
                None,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.NO_RESULT,
                {"added_tags": ["warning-1"], "removed_tags": []},
            ),
            (
                "s390x",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.IMPROVEMENT,
                {"added_tags": [], "removed_tags": ["error-1"]},
            ),
        ):
            if reference_result is not None:
                self.playground.advance_work_request(
                    reference_tasks[architecture],
                    result=reference_result,
                    use_data=False,
                )
            self.playground.advance_work_request(
                new_tasks[architecture], result=new_result, use_data=False
            )
            expected_analysis[architecture] = (status, details)

        self.assert_callback_status(
            new_lintian_workflow, WorkRequest.Statuses.PENDING
        )

        self.schedule_and_run_workflow_callback(new_lintian_workflow)

        new_lintian_workflow.refresh_from_db()
        self.assert_regression_analysis(
            new_lintian_workflow,
            "0.9-1",
            reference_results,
            "1.0-1",
            new_results,
            expected_analysis,
        )
        self.assert_callback_status(
            new_lintian_workflow, WorkRequest.Statuses.COMPLETED
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the signing no-operation task."""

from unittest import mock

from django.test import TestCase

import debusine.worker.tags as wtags
from debusine.artifacts.models import WorkRequestResults
from debusine.signing.tasks import SigningNoop
from debusine.signing.tasks.tests.helpers import SigningTaskTestCase
from debusine.tasks import TaskConfigError


class SigningNoopTests(SigningTaskTestCase[SigningNoop], TestCase):
    """Test the SigningNoop task."""

    SAMPLE_TASK_DATA = {"result": WorkRequestResults.SUCCESS}

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_get_subject(self) -> None:
        self.assertIsNone(self.get_subject())

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [wtags.TASK_PREFIX + "signing:noop:version:1"],
        )

    def test_execute_result_success(self) -> None:
        """If result=SUCCESS, the task returns SUCCESS."""
        with mock.patch.object(
            self.task, "_upload_work_request_debug_logs", autospec=True
        ):
            self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

    def test_execute_result_failure(self) -> None:
        """If result=FAILURE, the task returns FAILURE."""
        self.task_data["result"] = WorkRequestResults.FAILURE
        with mock.patch.object(
            self.task, "_upload_work_request_debug_logs", autospec=True
        ):
            self.assertEqual(self.task.execute(), WorkRequestResults.FAILURE)

    def test_no_additional_properties(self) -> None:
        """Assert no additional properties."""
        error_msg = "Extra inputs are not permitted"
        with self.assertRaisesRegex(TaskConfigError, error_msg):
            SigningNoop({"input": {}})



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for debusine.tests.django."""

import contextlib
import datetime as dt
import json
import re
from collections.abc import Callable, Generator
from pathlib import Path
from typing import Any, ClassVar, Never, Self, override
from unittest import mock

import django.template
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.db import transaction
from django.db.models import Model, QuerySet
from django.http.response import HttpResponse, HttpResponseBase, JsonResponse
from django.utils import timezone
from rest_framework import status
from rest_framework.response import Response

from debusine.db.context import context
from debusine.db.models import (
    ArtifactRelation,
    Scope,
    User,
    Workspace,
    permissions,
)
from debusine.db.models.permissions import Allow
from debusine.db.models.workspaces import WorkspaceQuerySet
from debusine.db.permissioncontext import PermissionContext, PermissionUser
from debusine.db.playground import scenarios
from debusine.server.views import ProblemResponse
from debusine.test.django import (
    AllowAll,
    ChannelsHelpersMixin,
    DenyAll,
    ListFilter,
    PermissionOverride,
    TestCase,
    override_permission,
)


class PermissionOverrideTests(TestCase):
    """Test permission overrides."""

    user: ClassVar[User]
    scope: ClassVar[Scope]
    scope1: ClassVar[Scope]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.scope = Scope.objects.get(name=settings.DEBUSINE_DEFAULT_SCOPE)
        cls.scope1 = Scope.objects.create(name="scope1")
        cls.user = get_user_model().objects.create_user(
            username="testuser", password="testpassword"
        )

    def assert_all_allowed(
        self, obj: PermissionOverride[Scope], user: PermissionUser
    ) -> None:
        """Check that all scopes test as allowed."""
        # Test allowed on permission check
        for scope in (self.scope, self.scope1):
            self.assertTrue(obj.check(scope, user))

        # Test allowed on permission filter
        self.assertQuerySetEqual(
            obj.filter(Scope.objects.all(), user),
            [self.scope, self.scope1],
            ordered=False,
        )

    def assert_allowed(
        self,
        obj: PermissionOverride[Scope],
        user: PermissionUser,
        scopes: list[Scope],
    ) -> None:
        """Check that all the given scopes test as allowed."""
        # Test allowed on permission check
        for scope in scopes:
            self.assertTrue(obj.check(scope, user))
        for scope in {self.scope, self.scope1} - set(scopes):
            self.assertFalse(obj.check(scope, user))

        # Test allowed on permission filter
        self.assertQuerySetEqual(
            obj.filter(Scope.objects.all(), user),
            scopes,
            ordered=False,
        )

    def test_allow_all(self) -> None:
        """Test AllowAll.check."""
        obj = AllowAll[Scope]()
        for user in (AnonymousUser(), self.user):
            with self.subTest(user=repr(user)):
                self.assert_all_allowed(obj, user)

    def test_deny_all_check(self) -> None:
        """Test DenyAll.check."""
        obj = DenyAll[Scope]()
        for user in (AnonymousUser(), self.user):
            with self.subTest(user=repr(user)):
                self.assert_allowed(obj, user, [])

    def test_list_noop(self) -> None:
        """Test ListFilter predicate with no arguments: allows all."""
        obj = ListFilter[Scope]()
        for user in (AnonymousUser(), self.user):
            with self.subTest(user=repr(user)):
                self.assert_all_allowed(obj, user)

    def test_list_include(self) -> None:
        """Test ListFilter with include."""
        obj = ListFilter[Scope](include=[self.scope])
        for user in (AnonymousUser(), self.user):
            with self.subTest(user=repr(user)):
                self.assert_allowed(obj, user, [self.scope])

    def test_list_exclude(self) -> None:
        """Test ListFilter with exclude."""
        obj = ListFilter[Scope](exclude=[self.scope])
        for user in (AnonymousUser(), self.user):
            with self.subTest(user=repr(user)):
                self.assert_allowed(obj, user, [self.scope1])

    def test_override_permission(self) -> None:
        """Test override_permission as context manager."""
        pc = PermissionContext.for_ui_user(self.user)
        with override_permission(Scope, "can_display", AllowAll):
            self.assertEqual(
                getattr(self.scope.can_display, "error_template"),
                "{user} cannot display scope {resource}",
            )
            self.assertTrue(self.scope.can_display(pc))
            self.assertQuerySetEqual(
                Scope.objects.can_display(pc),
                [self.scope, self.scope1],
                ordered=False,
            )
            self.assertQuerySetEqual(
                Scope.objects.can_display(pc),
                [self.scope, self.scope1],
                ordered=False,
            )
        with override_permission(Scope, "can_display", DenyAll):
            self.assertFalse(self.scope.can_display(pc))
            self.assertQuerySetEqual(Scope.objects.can_display(pc), [])
            self.assertQuerySetEqual(Scope.objects.can_display(pc), [])

    @override_permission(Scope, "can_display", DenyAll)
    def test_override_permission_decorator(self) -> None:
        """Test override_permission as decorator."""
        self.assertFalse(
            self.scope.can_display(PermissionContext.for_ui_user(self.user))
        )


class TestCaseTests(TestCase):
    """Tests for methods in debusine.test.django.TestCase."""

    playground_memory_file_store = False

    @contextlib.contextmanager
    def assertExceptionHasFormattedAPIResponse(
        self,
    ) -> Generator[None]:
        with mock.patch(
            "debusine.test.django.BaseDjangoTestCase.format_api_response",
            return_value="…",
        ):
            try:
                yield
            except Exception as exc:
                self.assertIn("…", exc.__notes__)
                raise

    def test_create_artifact_relation_default_type(self) -> None:
        """create_artifact_relation() create artifact with type=RELATED_TO."""
        artifact, _ = self.playground.create_artifact()
        target, _ = self.playground.create_artifact()
        created_artifact_relation = self.playground.create_artifact_relation(
            artifact, target
        )
        created_artifact_relation.refresh_from_db

        self.assertEqual(created_artifact_relation.artifact, artifact)
        self.assertEqual(created_artifact_relation.target, target)
        self.assertEqual(
            created_artifact_relation.type,
            ArtifactRelation.Relations.RELATES_TO,
        )

    def test_create_artifact_relation_specific_type(self) -> None:
        """create_artifact_relation() create artifact with given type."""
        artifact, _ = self.playground.create_artifact()
        target, _ = self.playground.create_artifact()
        relation_type = ArtifactRelation.Relations.BUILT_USING

        created_artifact_relation = self.playground.create_artifact_relation(
            artifact, target, relation_type
        )
        created_artifact_relation.refresh_from_db()

        self.assertEqual(created_artifact_relation.artifact, artifact)
        self.assertEqual(created_artifact_relation.target, target)
        self.assertEqual(created_artifact_relation.type, relation_type)

    def test_create_artifact_default_expire_date(self) -> None:
        """create_artifact() set expire_at to None by default."""
        artifact, _ = self.playground.create_artifact()
        artifact.refresh_from_db()
        self.assertIsNone(artifact.expire_at)

    def test_create_artifact_expire_date(self) -> None:
        """create_artifact() set expire_at use correct expire_at.."""
        now = timezone.now()
        artifact, _ = self.playground.create_artifact(expiration_delay=1)
        artifact.created_at = now - dt.timedelta(days=1)
        artifact.save()
        artifact.refresh_from_db()
        self.assertEqual(artifact.expire_at, now)

    def test_create_artifact_create_files(self) -> None:
        """create_artifact() returns an Artifact and create files."""
        paths = ["src/a", "b"]
        files_size = 12

        artifact, files_contents = self.playground.create_artifact(
            paths, files_size, create_files=True
        )

        self.assertEqual(artifact.fileinartifact_set.count(), len(paths))

        files_in_artifact = artifact.fileinartifact_set.order_by("path")

        for file_in_artifact in files_in_artifact:
            file = file_in_artifact.file
            file_backend = artifact.workspace.scope.download_file_backend(file)
            local_path = file_backend.get_local_path(file)
            assert local_path is not None
            self.assertEqual(
                local_path.read_bytes(), files_contents[file_in_artifact.path]
            )

    def test_create_artifact_do_not_create_files(self) -> None:
        """create_artifact() returns an Artifact and does not create files."""
        artifact, _ = self.playground.create_artifact(
            ["README"], create_files=False
        )

        self.assertEqual(artifact.fileinartifact_set.count(), 0)

    def test_create_artifact_raise_value_error(self) -> None:
        """create_artifact() raise ValueError: incompatible options."""
        with self.assertRaisesRegex(
            ValueError,
            "^skip_add_files_in_store must be False if create_files is False$",
        ):
            self.playground.create_artifact(
                create_files=False, skip_add_files_in_store=True
            )

    def test_format_api_response(self) -> None:
        for response, expected in (
            (
                HttpResponse("ok"),
                "200 OK (text/html; charset=utf-8): content='ok'",
            ),
            (
                HttpResponse(
                    "ok", status=status.HTTP_500_INTERNAL_SERVER_ERROR
                ),
                "500 Internal Server Error (text/html; charset=utf-8):"
                " content='ok'",
            ),
            (
                JsonResponse({"success": True}),
                "200 OK (application/json): content={'success': True}",
            ),
            (
                ProblemResponse(title="pr title"),
                "400 Bad Request (application/problem+json): title='pr title'",
            ),
            (
                ProblemResponse(title="", detail="pr detail"),
                "400 Bad Request (application/problem+json):"
                " title='' detail='pr detail'",
            ),
            (
                ProblemResponse(
                    title="", validation_errors={"arbitrary": "data"}
                ),
                "400 Bad Request (application/problem+json): title=''"
                " validation_errors={'arbitrary': 'data'}",
            ),
            (
                HttpResponse("x" * 300),
                f"200 OK (text/html; charset=utf-8): content='{'x' * 200}…'",
            ),
            (
                JsonResponse({"long": "x" * 200}),
                f"200 OK (application/json): content={{'long': '{'x' * 190}…",
            ),
            (
                HttpResponseBase(),
                "200 OK (text/html; charset=utf-8): content not available",
            ),
            (
                HttpResponseBase(content_type=None),
                "200 OK (text/html; charset=utf-8): content not available",
            ),
        ):
            with self.subTest(response=repr(response)):
                self.assertEqual(self.format_api_response(response), expected)

    def test_format_api_response_problemresponse_with_exception(self) -> None:
        response = ProblemResponse(title="Error")
        setattr(
            response, "_original_exception", Exception("original exception")
        )
        self.assertEqual(
            self.format_api_response(response),
            "400 Bad Request (application/problem+json):"
            " title='Error' original exception follows:\n"
            "Exception: original exception\n",
        )

    def test_assertAPIResponseOk_ok(self) -> None:
        data = {"success": True}
        response = JsonResponse(data)
        self.assertEqual(self.assertAPIResponseOk(response), data)

    def test_assertAPIResponseOk_not_json(self) -> None:
        response = HttpResponse("ok")
        with (
            self.assertRaisesRegex(
                AssertionError,
                r"content type 'text/html; charset=utf-8'"
                " is not a JSON response",
            ),
            self.assertExceptionHasFormattedAPIResponse(),
        ):
            self.assertAPIResponseOk(response)

    def test_assertAPIResponseOk_invalid_json(self) -> None:
        response = HttpResponse("{", content_type="application/json")
        with (
            self.assertRaisesRegex(
                json.JSONDecodeError,
                r"Expecting property name enclosed in double quotes",
            ),
            self.assertExceptionHasFormattedAPIResponse(),
        ):
            self.assertAPIResponseOk(response)

    def test_assertAPIResponseOk_bad_status(self) -> None:
        response = HttpResponse(
            "fail", status=status.HTTP_500_INTERNAL_SERVER_ERROR
        )
        with (
            self.assertRaisesRegex(AssertionError, r"500 != 200"),
            self.assertExceptionHasFormattedAPIResponse(),
        ):
            self.assertAPIResponseOk(response)

    def test_assertAPIResponseOk_no_content_type(self) -> None:
        response = HttpResponse("ok")
        del response.headers["Content-Type"]
        with (
            self.assertRaisesRegex(
                AssertionError, r"response has no content type"
            ),
            self.assertExceptionHasFormattedAPIResponse(),
        ):
            self.assertAPIResponseOk(response)

    def test_assertResponseProblem_valid(self) -> None:
        """assertResponseProblem() does not raise any exception."""
        title = "Invalid task name"
        response = ProblemResponse(title)

        self._process_json_response(response)

        self.assertResponseProblem(response, title)

    def test_assertResponseProblem_assertions(self) -> None:
        """Exercise all the checks done by assertResponseProblem()."""
        for response, title, detail_pattern, expected_regex in (
            (
                ProblemResponse("title", status_code=status.HTTP_200_OK),
                "title",
                "detail",
                "response status 200 != 400",
            ),
            (
                Response({}, status=status.HTTP_400_BAD_REQUEST),
                "title",
                None,
                r"application/problem\+json",
            ),
            (
                Response(
                    {},
                    status=status.HTTP_400_BAD_REQUEST,
                    content_type="something_invalid",
                ),
                "title",
                None,
                r"something_invalid",
            ),
            (
                ProblemResponse("actual-title"),
                "expected-title",
                None,
                'title "actual-title" != "expected-title"',
            ),
            (
                ProblemResponse("actual-title"),
                "actual-title",
                "expected-detail",
                '"detail" not found in ',
            ),
            (
                ProblemResponse("actual-title", "actual-detail"),
                "actual-title",
                "expected-detail",
                '"expected-detail" did not match "actual-detail"',
            ),
            (
                ProblemResponse("actual-title"),
                "actual-title",
                "expected-detail",
                '"detail" not found in response',
            ),
        ):
            with (
                self.subTest(
                    response=repr(response),
                    title=title,
                    detail_pattern=detail_pattern,
                ),
                self.assertRaisesRegex(self.failureException, expected_regex),
            ):
                response = response
                self._process_json_response(response)

                self.assertResponseProblem(
                    response, title, detail_pattern=detail_pattern
                )

    @staticmethod
    def _process_json_response(response: HttpResponse) -> None:
        """
        Simulate part of what Django does when returning responses.

        It adds a new method (json()) returning json.loads(response.content).
        If response.content_type exist assign it to
        response.headers["Content-Type"].
        """
        response.json = lambda: json.loads(  # type: ignore[attr-defined]
            response.content
        )

        if content_type := getattr(response, "content_type", False):
            response.headers["Content-Type"] = content_type

    def test_ephemeral_savepoint(self) -> None:
        """Test ephemeral_savepoint."""
        with self.ephemeral_savepoint():
            scope = self.playground.get_or_create_scope(
                "test_ephemeral_savepoint"
            )
        self.assertFalse(Scope.objects.filter(pk=scope.pk).exists())

    def test_ephemeral_savepoint_with_exception(self) -> None:
        """Test ephemeral_savepoint with exceptions."""
        with (
            self.assertRaisesRegex(RuntimeError, r"expected error"),
            self.ephemeral_savepoint() as sid,
        ):
            scope = self.playground.get_or_create_scope(
                "test_ephemeral_savepoint"
            )
            raise RuntimeError("expected error")
        self.assertTrue(Scope.objects.filter(pk=scope.pk).exists())
        transaction.savepoint_rollback(sid)
        self.assertFalse(Scope.objects.filter(pk=scope.pk).exists())

    def test_test_check_predicate_unscoped_resources(self) -> None:
        """Test checking permissions on unscoped resources."""
        scope = self.playground.get_default_scope()
        user = self.playground.get_default_user()
        pc = PermissionContext.for_ui_user(user)

        results = iter([True])
        with (
            mock.patch(
                "debusine.db.models.Scope.can_display",
                side_effect=lambda _: next(results),
            ),
            mock.patch(
                "debusine.test.django.get_resource_scope",
                return_value=None,
            ),
        ):
            self.assertEqual(
                self._test_check_predicate(pc, scope, "can_display", True),
                [],
            )

    def test_test_check_predicate_unscoped_resourcse(self) -> None:
        """Test checking permissions on non-workspaced resources."""
        scope = self.playground.get_default_scope()
        user = self.playground.get_default_user()
        pc = PermissionContext.for_ui_user(user)

        results = iter([True, True])
        with (
            mock.patch(
                "debusine.db.models.Scope.can_display",
                side_effect=lambda _: next(results),
            ),
        ):
            self.assertEqual(
                self._test_check_predicate(pc, scope, "can_display", True),
                [],
            )

    def test_test_check_predicate_fail_scope_in_context(self) -> None:
        """Test checking permissions failing with scope in context."""
        scope = self.playground.get_default_scope()
        user = self.playground.get_default_user()
        pc = PermissionContext.for_ui_user(user)

        results = iter([True, False])
        with (
            mock.patch(
                "debusine.db.models.Scope.can_display",
                side_effect=lambda _: next(results),
            ),
        ):
            self.assertEqual(
                self._test_check_predicate(pc, scope, "can_display", True),
                [
                    "can_display is False for playground on debusine"
                    " with scope set in context"
                ],
            )

    def test_test_check_predicate_fail_workspace_in_context(self) -> None:
        """Test checking permissions failing with scope in context."""
        workspace = self.playground.get_default_workspace()
        user = self.playground.get_default_user()
        pc = PermissionContext.for_ui_user(user)

        results = iter([True, True, False])
        with (
            mock.patch(
                "debusine.db.models.Workspace.can_display",
                side_effect=lambda _: next(results),
            ),
        ):
            self.assertEqual(
                self._test_check_predicate(pc, workspace, "can_display", True),
                [
                    "can_display is False for playground on debusine/System"
                    " with scope and workspace set in context"
                ],
            )

    def test_assertPermission_check_predicate(self) -> None:
        """Test _test_check_predicate."""
        public = self.playground.get_default_workspace()
        private = self.playground.create_workspace(name="private", public=False)
        self.assertPermission(
            "can_display",
            users=self.playground.get_default_user(),
            allowed=public,
            denied=private,
        )

    def test_assertPermission_collections(self) -> None:
        """Test passing collections to assertPermission."""
        public = self.playground.get_default_workspace()
        private = self.playground.create_workspace(name="private", public=False)
        self.assertPermission(
            "can_display",
            users=[self.playground.get_default_user()],
            allowed=(public,),
            denied={private},
        )

    def test_assertPermission_worker_token(self) -> None:
        """Test passing a worker token to assertPermission."""
        token = self.playground.create_worker_token()
        public = self.playground.get_default_workspace()
        private = self.playground.create_workspace(name="private", public=False)
        self.assertPermission(
            "can_display",
            users=None,
            denied=(public, private),
            token=token,
        )

    def test_assertPermission_work_request_token(self) -> None:
        """Test passing a work request token to assertPermission."""
        token = self.playground.create_work_request_token()
        public = self.playground.get_default_workspace()
        private = self.playground.create_workspace(name="private", public=False)
        self.assertPermission(
            "can_display",
            users=None,
            allowed=public,
            denied=private,
            token=token,
        )

    def test_assertPermission_user_token(self) -> None:
        """Test passing a user token to assertPermission."""
        token = self.playground.create_user_token()
        public = self.playground.get_default_workspace()
        private = self.playground.create_workspace(name="private", public=False)
        self.assertPermission(
            "can_display",
            users=self.playground.get_default_user(),
            allowed=public,
            denied=private,
            token=token,
        )

    def test_assertPermission_external_user_token(self) -> None:
        """Test passing an external user token to assertPermission."""
        public = self.playground.get_default_workspace()
        private = self.playground.create_workspace(name="private", public=False)
        token = self.playground.create_external_user_token()
        self.assertPermission(
            "can_display",
            users=AnonymousUser(),
            denied=(public, private),
            token=token,
        )

    def test_assertPermission_fails(self) -> None:
        """Test assertPermission failure reporting."""
        context.set_permission_context_scoped(
            self.playground.get_default_scope(),
            PermissionContext.for_ui_user(self.playground.get_default_user()),
        )
        public = self.playground.get_default_workspace()
        private = self.playground.create_workspace(name="private", public=False)

        with self.assertRaises(self.failureException) as exc:
            self.assertPermission(
                "can_display",
                users=[self.playground.get_default_user()],
                allowed=(private,),
                denied={public},
            )
        self.assertEqual(
            exc.exception.args[0],
            "Predicate permission mismatch:\n"
            "* can_display is False for playground on debusine/private"
            " with empty context\n"
            "* can_display is False for playground on debusine/private"
            " with scope set in context\n"
            "* can_display is True for playground on debusine/System"
            " with empty context\n"
            "* can_display is True for playground on debusine/System"
            " with scope set in context\n"
            "* can_display is True for playground on debusine/System"
            " with scope and workspace set in context\n"
            "* can_display for playground selects debusine/System"
            " instead of debusine/private\n"
            "* can_display for playground denies debusine/private"
            " instead of debusine/System\n",
        )

        with self.assertRaises(self.failureException) as exc:
            self.assertPermission(
                "can_display",
                users=[self.playground.get_default_user()],
                allowed=(),
                denied=(private,),
            )
        self.assertEqual(
            exc.exception.args[0],
            "Predicate permission mismatch:\n"
            "* can_display for playground selects debusine/System"
            " instead of nothing\n",
        )

    def test_assertPermission_fails1(self) -> None:
        """Test assertPermission failure reporting."""
        workspace = self.playground.get_default_workspace()
        workspace.public = False
        workspace.save()

        with self.assertRaises(self.failureException) as exc:
            self.assertPermission(
                "can_display",
                users=[self.playground.get_default_user()],
                allowed=(workspace),
            )
        self.assertEqual(
            exc.exception.args[0],
            "Predicate permission mismatch:\n"
            "* can_display is False for playground on debusine/System"
            " with empty context\n"
            "* can_display is False for playground on debusine/System"
            " with scope set in context\n"
            "* can_display for playground selects nothing"
            " instead of debusine/System\n"
            "* can_display for playground denies debusine/System"
            " instead of nothing\n",
        )

    def test_assertPermission_denied(self) -> None:
        """Test assertPermission failure reporting."""
        workspace = self.playground.get_default_workspace()
        workspace.public = False
        workspace.save()
        workspace1 = self.playground.create_workspace(
            name="other", public=False
        )

        with self.assertRaises(self.failureException) as exc:
            self.assertPermission(
                "can_display",
                users=[self.playground.get_default_user()],
                denied=[workspace],
            )
        self.assertEqual(
            exc.exception.args[0],
            "Predicate permission mismatch:\n"
            f"* can_display for playground denies {workspace}, {workspace1}"
            f" instead of {workspace}\n",
        )

    def test_check_predicate_filter_output(self) -> None:
        for prefix, filtered, expected, message in (
            ("has ", [], [], None),
            ("has ", ["a"], ["a"], None),
            ("has ", ["a"], [], "has a instead of nothing"),
            ("has ", [], ["a"], "has nothing instead of a"),
            ("has ", ["a", "b"], ["1", "2"], "has a, b instead of 1, 2"),
        ):
            with self.subTest(
                prefix=prefix, filtered=filtered, expected=expected
            ):
                errors: list[str] = []
                self._check_predicate_filter_output(
                    prefix, filtered, expected, errors
                )
                if message is None:
                    self.assertEqual(errors, [])
                else:
                    self.assertEqual(errors, [message])

    def test_build_permission_when_role_testlist(self) -> None:
        """Test the _build_permission_when_role_testlist method."""
        scope = self.playground.get_or_create_scope("scope")
        workspace = self.playground.create_workspace(
            name="workspace", scope=scope
        )
        # Alias the long function name
        f = self._build_permission_when_role_testlist
        self.assertEqual(
            f(scope.can_display, "owner", scope_roles="owner"),
            [
                (scope, Scope.Roles.OWNER),
                (scope, Scope.Roles.OWNER),
            ],
        )
        self.assertEqual(
            f(
                workspace.can_display,
                [Workspace.Roles.OWNER, "contributor"],
                scope_roles="owner",
                workspace_roles="owner",
            ),
            [
                (scope, Scope.Roles.OWNER),
                (workspace, Workspace.Roles.OWNER),
                (workspace, Workspace.Roles.OWNER),
                (workspace, Workspace.Roles.CONTRIBUTOR),
            ],
        )
        with self.assertRaises(AssertionError):
            f(scope.can_display, workspace_roles="owner")

    def test_assertPermissionWhenRole(self) -> None:
        """Test assertPermissionWhenRole."""
        scope = self.playground.get_or_create_scope("scope")
        workspace = self.playground.create_workspace(
            name="workspace", scope=scope
        )
        user = self.playground.get_default_user()
        pc = PermissionContext.for_ui_user(user)
        work_request = self.playground.create_worker_task(workspace=workspace)
        self.assertPermissionWhenRole(
            scope.can_create_workspace, pc, roles="owner"
        )
        self.assertPermissionWhenRole(
            scope.can_create_workspace, pc, scope_roles="owner"
        )
        self.assertPermissionWhenRole(
            work_request.can_unblock, pc, workspace_roles="owner"
        )
        self.assertNoPermissionWhenRole(
            work_request.can_unblock, pc, workspace_roles="contributor"
        )
        with self.assertRaisesRegex(
            self.failureException, r"is false even with contributor"
        ):
            self.assertPermissionWhenRole(
                work_request.can_unblock, pc, workspace_roles="contributor"
            )
        with self.assertRaisesRegex(
            self.failureException, r"is true when given owner"
        ):
            self.assertNoPermissionWhenRole(
                scope.can_create_workspace, pc, roles="owner"
            )

    def test_assertPermissionWhenRole_had_perm(self) -> None:
        """Test assertPermissionWhenRole when permission was already there."""
        scope = self.playground.get_default_scope()
        user = self.playground.get_default_user()
        pc = PermissionContext.for_ui_user(user)
        self.playground.create_group_role(
            scope, Scope.Roles.OWNER, users=[user]
        )
        with self.assertRaisesRegex(
            self.failureException, r"is true even without owner"
        ):
            self.assertPermissionWhenRole(scope.can_display, pc, "owner")
        with self.assertRaisesRegex(
            self.failureException, r"is true even without owner"
        ):
            self.assertNoPermissionWhenRole(scope.can_display, pc, "owner")

    def test_custom_template(self) -> None:
        """Test installing a custom template."""
        with self.custom_template("web/foo.html") as path:
            self.assertEqual(
                path,
                Path(settings.DEBUSINE_TEMPLATE_DIRECTORY) / "web" / "foo.html",
            )
            path.write_text("TEST")

        self.assertFalse(path.exists())

    def test_custom_template_not_written(self) -> None:
        """Test installing a custom template."""
        with self.custom_template("web/foo.html") as path:
            pass
        self.assertFalse(path.exists())

    def test_clear_template_cache(self) -> None:
        """Test clearing the template cache."""
        engine = django.template.engines["django"]

        with (
            self.custom_template("main.html") as main,
            self.custom_template("included.html") as included,
        ):
            main.write_text("A{% include 'included.html' %}")
            included.write_text("B")
            template = engine.get_template("main.html")
            self.assertEqual(template.render({}), "AB")

            # Changing the contents of an included template doesn't invalidate
            # the template cache
            included.write_text("C")
            template = engine.get_template("main.html")
            self.assertEqual(template.render({}), "AB")

            # Invalidating it manually works
            self.clear_template_cache()
            template = engine.get_template("main.html")
            self.assertEqual(template.render({}), "AC")

    def test_clear_template_cache_skip_unknown(self) -> None:
        """Test skipping unknown engines and loaders."""
        from django.template.backends.django import DjangoTemplates
        from django.template.loaders.base import Loader

        mock_django_loader = mock.Mock(spec=Loader)
        mock_other_loader = mock.Mock()
        mock_loaders = [mock_django_loader, mock_other_loader]

        mock_django_engine = mock.Mock(spec=DjangoTemplates)
        mock_django_engine.engine = mock.Mock()
        mock_django_engine.engine.template_loaders = mock_loaders
        mock_other_engine = mock.Mock()
        mock_engines = [mock_django_engine, mock_other_engine]

        with mock.patch(
            "django.template.engines.all", return_value=mock_engines
        ):
            self.clear_template_cache()

        mock_django_loader.reset.assert_called()
        mock_other_loader.reset.assert_not_called()


class TestChannelsHelpersMixinTests(ChannelsHelpersMixin, TestCase):
    """Tests for methods in ChannelsHelpersMixin."""

    # Default channel name to be used during the tests.
    channel_name = "generic-channel-for-testing"

    async def test_create_channel(self) -> None:
        """Create channel return a dictionary with layer and name keys."""
        channel = await self.create_channel("channel-test")

        self.assertEqual(channel.keys(), {"layer", "name"})

    async def test_assert_channel_received_raises_exception(self) -> None:
        """assert_channel_received raise exception: nothing was received."""
        channel = await self.create_channel(self.channel_name)
        with self.assertRaisesRegex(
            self.failureException,
            "^Expected '{'type': 'work_request'}' received nothing$",
        ):
            await self.assert_channel_received(
                channel, {"type": "work_request"}
            )

    async def test_assert_channel_received_raise_wrong_data(self) -> None:
        """assert_channel_received raise exception: unexpected data received."""
        channel = await self.create_channel(self.channel_name)
        message = {"type": "work_request.assigned"}
        await channel["layer"].group_send(
            self.channel_name, {"some other message": "values"}
        )

        with self.assertRaises(AssertionError):
            await self.assert_channel_received(channel, message)

    async def test_assert_channel_received_do_not_raise(self) -> None:
        """assert_channel_received does not raise an exception."""
        channel = await self.create_channel(self.channel_name)
        message = {"type": "work_request.assigned"}
        await channel["layer"].group_send(self.channel_name, message)

        await self.assert_channel_received(channel, message)

    async def test_assert_channel_nothing_received_do_not_raise(self) -> None:
        """assert_channel_nothing_received does not raise an exception."""
        channel = await self.create_channel(self.channel_name)
        await self.assert_channel_nothing_received(channel)

    async def test_assert_channel_nothing_receive_raise(self) -> None:
        """assert_channel_nothing_received raise exception: data is received."""
        channel = await self.create_channel(self.channel_name)
        message = {"type": "work_request.assigned"}

        await channel["layer"].group_send(self.channel_name, message)

        with self.assertRaisesRegex(
            self.failureException,
            "^Expected nothing. Received: '{'type': 'work_request.assigned'}'$",
        ):
            await self.assert_channel_nothing_received(channel)


class AssertRolePredicateTests(TestCase):
    """Tests for assertRolePredicate."""

    scenario = scenarios.DefaultContext()

    def assertFails(
        self, pattern: str | re.Pattern[str], resource: Model | None = None
    ) -> None:
        """Call the test and expect it to fail with matching error."""
        if resource is None:
            resource = self.scenario.workspace
        if isinstance(pattern, str):
            pattern = re.compile(re.escape(pattern))
        with self.assertRaisesRegex(self.failureException, pattern):
            self.assertRolePredicate(
                resource,
                "can_configure",
                Workspace.Roles.OWNER,
                work_request=Allow.NEVER,
            )

    def test_no_perm_check(self) -> None:
        self.assertFails(
            "object has no @permission_check can_configure",
            resource=object(),  # type:ignore[arg-type]
        )

    def test_perm_check_undecorated(self) -> None:
        def f(_: PermissionUser | None) -> Never:
            raise NotImplementedError()

        with mock.patch("debusine.db.models.Workspace.can_configure", f):
            self.assertFails(
                "Workspace.can_configure is not marked as @permission_check"
            )

    def test_perm_check_wrong_result(self) -> None:
        def _pred(
            result: bool,
        ) -> Callable[[Workspace, PermissionContext], bool]:
            @permissions.permission_check("fail", work_request=Allow.NEVER)
            def _predicate(self: Workspace, pc: PermissionContext) -> bool:
                self.has_role(pc, Workspace.Roles.OWNER)
                return result

            return _predicate

        with mock.patch(
            "debusine.db.models.Workspace.can_configure", _pred(True)
        ):
            self.assertFails("True is not false")
        with mock.patch(
            "debusine.db.models.Workspace.can_configure", _pred(False)
        ):
            self.assertFails("False != True")

    def test_perm_check_wrong_role(self) -> None:
        def _pred(
            *roles: Workspace.Roles,
        ) -> Callable[[Workspace, PermissionContext], bool]:
            results = list(roles)

            @permissions.permission_check("fail", work_request=Allow.NEVER)
            def _predicate(self: Workspace, pc: PermissionContext) -> bool:
                return self.has_role(pc, results.pop(0))

            return _predicate

        with mock.patch(
            "debusine.db.models.Workspace.can_configure",
            _pred(Workspace.Roles.CONTRIBUTOR),
        ):
            self.assertFails("expected call not found")

        with mock.patch(
            "debusine.db.models.Workspace.can_configure",
            _pred(Workspace.Roles.OWNER, Workspace.Roles.CONTRIBUTOR),
        ):
            self.assertFails("expected call not found")

    def test_no_manager_filter(self) -> None:
        with mock.patch("debusine.db.models.Workspace.objects", object()):
            self.assertFails(
                "Workspace.objects has no @permission_filter can_configure"
            )

    def test_manager_filter_undecorated(self) -> None:
        def f(_: PermissionUser | None) -> Never:
            raise NotImplementedError()

        with mock.patch.object(Workspace.objects, "can_configure", f):
            self.assertFails(
                "Workspace.objects.can_configure"
                " is not marked as @permission_filter"
            )

    def test_manager_filter_wrong_result(self) -> None:
        @permissions.permission_filter(work_request=Allow.NEVER)
        def _predicate(
            self: WorkspaceQuerySet[Any], pc: PermissionContext
        ) -> WorkspaceQuerySet[Any]:
            self.with_role(pc, Workspace.Roles.OWNER)
            return None  # type: ignore[return-value]

        with mock.patch(
            "debusine.db.models.workspaces.WorkspaceQuerySet.can_configure",
            _predicate,
        ):
            self.assertFails(
                re.compile("None is not an instance of <class.+QuerySet")
            )

    def test_manager_filter_wrong_role(self) -> None:
        @permissions.permission_filter(work_request=Allow.NEVER)
        def _predicate(
            self: WorkspaceQuerySet[Any], pc: PermissionContext
        ) -> WorkspaceQuerySet[Any]:
            return self.with_role(pc, Workspace.Roles.CONTRIBUTOR)

        with mock.patch(
            "debusine.db.models.workspaces.WorkspaceQuerySet.can_configure",
            _predicate,
        ):
            self.assertFails("expected call not found")

    def test_no_queryset_filter(self) -> None:
        class TestQuerySet(QuerySet[Workspace]):
            def with_role(
                self,
                user: PermissionUser,  # noqa: U100
                role: Workspace.Roles,  # noqa: U100
            ) -> Self:
                raise NotImplementedError()

        with mock.patch(
            "debusine.db.models.workspaces.WorkspaceManager.all",
            return_value=TestQuerySet(Workspace),
        ):
            self.assertFails(
                "QuerySet[Workspace] has no @permission_filter can_configure"
            )

    def test_queryset_filter_undecorated(self) -> None:
        class TestQuerySet(QuerySet[Workspace]):
            def with_role(
                self,
                user: PermissionUser,
                role: Workspace.Roles,  # noqa: U100
            ) -> Self:
                raise NotImplementedError()

            def can_configure(self, _: PermissionUser | None) -> Self:
                raise NotImplementedError()

        with mock.patch(
            "debusine.db.models.workspaces.WorkspaceManager.all",
            return_value=TestQuerySet(Workspace),
        ):
            self.assertFails(
                "QuerySet[Workspace].can_configure"
                " is not marked as @permission_filter"
            )

    def test_queryset_filter_wrong_result(self) -> None:
        class TestQuerySet(QuerySet[Workspace]):
            def with_role(
                self,
                pc: PermissionContext,
                role: Workspace.Roles,  # noqa: U100
            ) -> Self:
                raise NotImplementedError()

            @permissions.permission_filter(work_request=Allow.NEVER)
            def can_configure(self, pc: PermissionContext) -> Self:
                self.with_role(pc, Workspace.Roles.OWNER)
                return None  # type: ignore[return-value]

        with mock.patch(
            "debusine.db.models.workspaces.WorkspaceManager.all",
            return_value=TestQuerySet(Workspace),
        ):
            self.assertFails(
                re.compile("None is not an instance of <class.+QuerySet")
            )

    def test_queryset_filter_wrong_role(self) -> None:
        class TestQuerySet(WorkspaceQuerySet[Any]):
            @override
            def with_role(
                self,
                pc: PermissionContext,
                role: Workspace.Roles,  # noqa: U100
            ) -> Self:
                raise NotImplementedError()

            @override
            @permissions.permission_filter(work_request=Allow.NEVER)
            def can_configure(self, pc: PermissionContext) -> Self:
                return self.with_role(pc, Workspace.Roles.CONTRIBUTOR)

        with mock.patch(
            "debusine.db.models.workspaces.WorkspaceManager.all",
            return_value=TestQuerySet(Workspace),
        ):
            self.assertFails("expected call not found")

    def test_assertRolePredicateResult_expect_false(self) -> None:
        @permissions.permission_check("fail")
        def _check(self: Workspace, pc: PermissionContext) -> bool:
            self.has_role(pc, Workspace.Roles.OWNER)
            return False

        class _QuerySet(WorkspaceQuerySet[Any]):
            @override
            @permissions.permission_filter(work_request=Allow.NEVER)
            def can_configure(self, pc: PermissionContext) -> Self:
                return self.with_role(pc, Workspace.Roles.OWNER).none()

        with (
            mock.patch(
                "debusine.db.models.workspaces.Workspace.can_configure",
                _check,
            ),
            mock.patch(
                "debusine.db.models.workspaces.WorkspaceManager.get_queryset",
                return_value=_QuerySet(Workspace),
            ),
            mock.patch(
                "debusine.db.models.workspaces.WorkspaceManager.all",
                return_value=_QuerySet(Workspace),
            ),
        ):
            self.assertRolePredicateResult(
                self.scenario.workspace,
                "can_configure",
                Workspace.Roles.OWNER,
                pc=PermissionContext.for_ui_user(self.scenario.user),
                expected_when_has_role=False,
            )

    def test_predicate(self) -> None:
        self.assertRolePredicate(
            self.scenario.workspace,
            "can_configure",
            Workspace.Roles.OWNER,
            work_request=Allow.NEVER,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for HeadersMiddleware."""

from django.http import HttpResponse
from django.test import RequestFactory
from django.utils.cache import has_vary_header

from debusine.server.middlewares.headers import HeadersMiddleware
from debusine.test.django import TestCase


class HeadersMiddlewareTests(TestCase):
    """Test HeadersMiddleware."""

    def test_unauthenticated(self) -> None:
        wrapped_view = HeadersMiddleware(lambda _: HttpResponse())
        response = wrapped_view(RequestFactory().get("/"))
        assert isinstance(response, HttpResponse)
        self.assertTrue(has_vary_header(response, "Cookie"))
        self.assertTrue(has_vary_header(response, "Token"))

    def test_authenticated_with_cookie(self) -> None:
        wrapped_view = HeadersMiddleware(lambda _: HttpResponse())
        response = wrapped_view(
            RequestFactory().get("/", headers={"Cookie": "tasty"})
        )
        assert isinstance(response, HttpResponse)
        self.assertTrue(has_vary_header(response, "Cookie"))
        self.assertTrue(has_vary_header(response, "Token"))

    def test_authenticated_with_token(self) -> None:
        wrapped_view = HeadersMiddleware(lambda _: HttpResponse())
        response = wrapped_view(
            RequestFactory().get("/", headers={"Token": "valid"})
        )
        assert isinstance(response, HttpResponse)
        self.assertTrue(has_vary_header(response, "Cookie"))
        self.assertTrue(has_vary_header(response, "Token"))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for application context."""

import asyncio
from collections.abc import Callable
from threading import Thread
from typing import Any, ClassVar, NamedTuple, Self, override

from django.db import connection, transaction

from debusine.client.models import LookupChildType
from debusine.db.context import ContextConsistencyError, context
from debusine.db.models import Scope, Token, Workspace
from debusine.db.models.auth import system_user
from debusine.db.models.lookups import (
    LookupCache,
    LookupCacheGroup,
    LookupResultBare,
)
from debusine.db.models.scopes import ScopeRoles
from debusine.db.models.workspaces import WorkspaceRoles
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.test.django import (
    PlaygroundTestCase,
    TestCase,
    TransactionTestCase,
)


class ContextSnapshot(NamedTuple):
    """Comparable snapshot of the contents of a context."""

    scope: Scope | None
    pc: PermissionContext | None
    scope_roles: frozenset[ScopeRoles] | None
    workspace: Workspace | None
    workspace_roles: frozenset[WorkspaceRoles] | None
    lookup_cache_sizes: tuple[int, int] | None

    @classmethod
    def take(cls) -> Self:
        """Take a snapshot of the contents of a context."""
        lookup_cache = context._lookup_cache.get()
        return cls(
            scope=context._scope.get(),
            pc=context._permission_context.get(),
            scope_roles=context._scope_roles.get(),
            workspace=context._workspace.get(),
            workspace_roles=context._workspace_roles.get(),
            lookup_cache_sizes=(
                (
                    len(lookup_cache.single),
                    len(lookup_cache.dictionary),
                )
                if lookup_cache
                else None
            ),
        )


class TestContextBase(PlaygroundTestCase):
    """Common base for context tests."""

    scenario = scenarios.DefaultContext()

    @override
    def setUp(self) -> None:
        super().setUp()
        self.lookup_cache_single_key = (
            1,
            self.scenario.workspace,
            self.scenario.user,
            None,
            None,
            LookupChildType.ANY,
        )
        self.lookup_cache_single_value = LookupResultBare(lookup="")

    def assert_context_initial(self) -> None:
        """Check that the context has initial values."""
        with self.assertRaisesRegex(
            ContextConsistencyError, r"permission context has not yet been set"
        ):
            context.user_token
        self.assertIsNone(context.scope)
        self.assertIsNone(context._scope_roles.get())
        self.assertIsNone(context.workspace)
        self.assertIsNone(context._workspace_roles.get())
        self.assertIsNone(context._permission_context.get())
        self.assertFalse(context.permission_checks_disabled)
        self.assertEqual(context.lookup_cache, LookupCache())
        with self.assertRaisesRegex(
            ContextConsistencyError, r"scope is not set in context"
        ):
            context.require_scope()
        with self.assertRaisesRegex(
            ContextConsistencyError, r"permission context has not yet been set"
        ):
            context.user
        with self.assertRaisesRegex(
            ContextConsistencyError, r"workspace is not set in context"
        ):
            context.require_workspace()

    def assert_context_all_set(self) -> None:
        """Check that the context has been set to default test values."""
        self.assertEqual(
            context._permission_context.get(), self.scenario.context_ui()
        )
        self.assertEqual(context.user, self.scenario.user)
        self.assertEqual(context.scope, self.scenario.scope)
        self.assertEqual(context.require_scope(), self.scenario.scope)
        self.assertIsNotNone(context._scope_roles.get())
        self.assertEqual(context.workspace, self.scenario.workspace)
        self.assertIsNotNone(context._workspace_roles.get())
        self.assertEqual(
            context.require_workspace(),
            self.scenario.workspace,
        )
        self.assertEqual(
            context.lookup_cache,
            LookupCache(
                single=LookupCacheGroup(
                    {
                        self.lookup_cache_single_key: (
                            self.lookup_cache_single_value
                        )
                    }
                )
            ),
        )


class TestAppContext(TestContextBase, TestCase):
    """Test application context variables."""

    user_token: ClassVar[Token]
    worker_token: ClassVar[Token]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user_token = cls.playground.create_user_token()
        cls.worker_token = cls.playground.create_worker_token()

    @override
    def setUp(self) -> None:
        """Ensure context is not left dirty by a bug in a previous test."""
        super().setUp()
        context.reset()
        self.assertIsNone(context.scope)
        self.assertIsNone(context.workspace)
        self.assertIsNone(context._permission_context.get())

    def run_in_task(
        self, func: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> Any:
        """Run a callable in an asyncio task."""

        async def _task_main() -> Any:
            return func(*args, **kwargs)

        async def _async_main() -> asyncio.Task[Any]:
            return await asyncio.create_task(_task_main())

        return asyncio.run(_async_main())

    def test_defaults(self) -> None:
        """Context variables are None by default."""
        self.assert_context_initial()

    def test_local_previously_none(self) -> None:
        """Using local() restores previously unset vars."""
        with context.local():
            self.scenario.set_current()
            context.lookup_cache.single[self.lookup_cache_single_key] = (
                self.lookup_cache_single_value
            )
            self.assert_context_all_set()

        self.assert_context_initial()

    def test_local_previously_set(self) -> None:
        """Using local() restores previously set vars."""
        try:
            self.scenario.set_current()
            context.lookup_cache.single[self.lookup_cache_single_key] = (
                self.lookup_cache_single_value
            )

            with context.local():
                context.reset()
                self.assert_context_initial()

            self.assert_context_all_set()
        finally:
            context.reset()

    def test_reset(self) -> None:
        """Calling reset() sets context to its initial values."""
        self.scenario.set_current()
        context.lookup_cache.single[self.lookup_cache_single_key] = (
            self.lookup_cache_single_value
        )
        self.assert_context_all_set()

        context.reset()
        self.assert_context_initial()

    def test_visibility_task(self) -> None:
        """Check visibility with asyncio tasks."""
        self.scenario.set_current()
        context.lookup_cache.single[self.lookup_cache_single_key] = (
            self.lookup_cache_single_value
        )

        def _test() -> None:
            # Application context is preserved when entering a Task
            self.assert_context_all_set()
            context.reset()
            self.assert_context_initial()

        self.run_in_task(_test)

        # Changes made in another Task are isolated
        self.assert_context_all_set()

    def test_set_permission_context_scoped(self) -> None:
        self.playground.create_group_role(
            self.scenario.scope, Scope.Roles.OWNER, users=[self.scenario.user]
        )

        self.assert_context_initial()
        context.set_permission_context_scoped(
            self.scenario.scope, self.scenario.context_ui()
        )
        self.assertEqual(context.scope, self.scenario.scope)
        self.assertEqual(context.pc, self.scenario.context_ui())
        self.assertIsNone(context.workspace)
        self.assertIsNone(context._workspace_roles.get())
        self.assertFalse(context.permission_checks_disabled)

        self.assertEqual(context.scope_roles, frozenset((Scope.Roles.OWNER,)))

    def test_set_permission_context_scoped_extra_groups(self) -> None:
        pc = self.playground.create_pc_with_extra_groups(
            self.scenario.scope_owners
        )
        self.assert_context_initial()
        context.set_permission_context_scoped(self.scenario.scope, pc)
        self.assertEqual(context.scope, self.scenario.scope)
        self.assertEqual(context.pc, pc)
        self.assertIsNone(context.workspace)
        self.assertIsNone(context._workspace_roles.get())
        self.assertFalse(context.permission_checks_disabled)

        self.assertEqual(context.scope_roles, frozenset((Scope.Roles.OWNER,)))

    def test_set_permission_context_scoped_twice(self) -> None:
        context.set_permission_context_scoped(
            self.scenario.scope, self.scenario.context_ui()
        )
        expected = ContextSnapshot.take()

        scope1 = self.playground.get_or_create_scope(name="scope1")
        pc1 = PermissionContext.for_ui_anonymous()
        with self.assertRaisesRegex(
            ContextConsistencyError, "Scope was already set to debusine"
        ):
            context.set_permission_context_scoped(scope1, pc1)

        self.assertEqual(ContextSnapshot.take(), expected)

    def test_set_permission_context_scopeless(self) -> None:
        self.assert_context_initial()
        context.set_permission_context_scopeless(self.scenario.context_ui())
        self.assertIsNone(context.scope)
        self.assertEqual(context.pc, self.scenario.context_ui())
        with self.assertRaisesRegex(
            ContextConsistencyError, "scope has not been set"
        ):
            self.assertIsNone(context.scope_roles)
        self.assertIsNone(context.workspace)
        self.assertIsNone(context._workspace_roles.get())
        self.assertFalse(context.permission_checks_disabled)

    def test_set_permission_context_scopeless_twice(self) -> None:
        context.set_permission_context_scopeless(self.scenario.context_ui())
        expected = ContextSnapshot.take()

        pc1 = PermissionContext.for_ui_anonymous()
        with self.assertRaisesRegex(
            ContextConsistencyError, "Permission context was already set to "
        ):
            context.set_permission_context_scopeless(pc1)

        self.assertEqual(ContextSnapshot.take(), expected)

    def test_set_permission_context_scopeless_then_scoped(self) -> None:
        context.set_permission_context_scopeless(self.scenario.context_ui())
        expected = ContextSnapshot.take()

        with self.assertRaisesRegex(
            ContextConsistencyError, "Permission context was already set to "
        ):
            context.set_permission_context_scoped(
                self.scenario.scope, self.scenario.context_ui()
            )

        self.assertEqual(ContextSnapshot.take(), expected)

    def test_set_permission_context_scoped_then_scopeless(self) -> None:
        context.set_permission_context_scoped(
            self.scenario.scope, self.scenario.context_ui()
        )
        expected = ContextSnapshot.take()

        with self.assertRaisesRegex(
            ContextConsistencyError,
            f"Scope was already set to {self.scenario.scope}",
        ):
            context.set_permission_context_scopeless(self.scenario.context_ui())

        self.assertEqual(ContextSnapshot.take(), expected)

    def test_scope_roles_unset(self) -> None:
        """Test scope_roles accessor."""
        with self.assertRaisesRegex(
            ContextConsistencyError, "scope has not been set"
        ):
            context.scope_roles

    def test_scope_roles_empty(self) -> None:
        """Test scope_roles accessor with empty roleset."""
        self.scenario.set_current()
        self.assertEqual(context.scope_roles, frozenset())

    def test_scope_roles(self) -> None:
        """Test scope_roles accessor."""
        self.playground.create_group_role(
            self.scenario.scope, Scope.Roles.OWNER, users=[self.scenario.user]
        )
        self.scenario.set_current()
        self.assertEqual(context.scope_roles, frozenset((Scope.Roles.OWNER,)))

    def test_workspace_roles_unset(self) -> None:
        """Test workspace_roles accessor."""
        with self.assertRaisesRegex(
            ContextConsistencyError, "workspace is not set"
        ):
            context.workspace_roles

    def test_workspace_roles_empty(self) -> None:
        """Test workspace_roles accessor with empty roleset."""
        self.scenario.set_current()
        self.assertEqual(
            context.workspace_roles, frozenset((Workspace.Roles.VIEWER,))
        )

    def test_workspace_roles(self) -> None:
        """Test workspace_roles accessor."""
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.OWNER,
            users=[self.scenario.user],
        )

        self.scenario.set_current()
        self.assertEqual(
            context.workspace_roles, frozenset((Scope.Roles.OWNER,))
        )

    def test_management_command(self) -> None:
        self.scenario.set_current()

        with context.management_command():
            self.assertIsNone(context.scope)
            self.assertIsNone(context._scope_roles.get())
            self.assertIsNone(context.workspace)
            self.assertIsNone(context._workspace_roles.get())
            self.assertFalse(context.permission_checks_disabled)
            self.assertEqual(context.lookup_cache, LookupCache())
            self.assertEqual(context.user, system_user())

        self.assertIsNone(context.user_token)
        self.assertEqual(context.user, self.scenario.user)
        self.assertEqual(context.scope, self.scenario.scope)
        self.assertEqual(context.require_scope(), self.scenario.scope)
        self.assertIsNotNone(context._scope_roles.get())
        self.assertEqual(context.workspace, self.scenario.workspace)
        self.assertIsNotNone(context._workspace_roles.get())
        self.assertEqual(
            context.require_workspace(),
            self.scenario.workspace,
        )
        self.assertFalse(context.permission_checks_disabled)
        self.assertEqual(context.lookup_cache, LookupCache())


class TestThread(TestContextBase, TransactionTestCase):
    """Test thread isolation for application context."""

    worker_token: Token

    @override
    def setUp(self) -> None:
        super().setUp()
        with transaction.atomic():
            self.worker_token = self.playground.create_worker_token()

    def run_in_thread(
        self, func: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> Any:
        """Run a callable in a thread."""
        result: Any = None

        def _thread_main() -> None:
            nonlocal result
            try:
                result = func(*args, **kwargs)
            finally:
                connection.close()

        thread = Thread(target=_thread_main)
        thread.start()
        thread.join()

        return result

    def test_visibility_thread(self) -> None:
        """Check visibility with subthreads."""
        scope1 = self.playground.get_or_create_scope(name="scope1")
        workspace1 = self.playground.create_workspace(
            scope=scope1, name="test1", public=True
        )
        user1 = self.playground.create_user("test1")

        self.assert_context_initial()
        self.scenario.set_current()
        context.lookup_cache.single[self.lookup_cache_single_key] = (
            self.lookup_cache_single_value
        )

        def _test() -> None:
            # Application context is cleared when changing Thread
            self.assert_context_initial()

            context.set_permission_context_scoped(
                scope1, PermissionContext.for_ui_user(user1)
            )
            workspace1.set_current()
            context.lookup_cache.single[self.lookup_cache_single_key] = (
                self.lookup_cache_single_value
            )

            self.assertEqual(context.scope, scope1)
            self.assertIsNotNone(context._scope_roles.get())
            self.assertEqual(context.user, user1)
            self.assertEqual(context.workspace, workspace1)
            self.assertIsNotNone(context._workspace_roles.get())

        self.run_in_thread(_test)

        self.assert_context_all_set()



# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import datetime as dt
import os
import sys

import django
from django.conf import settings

sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('_ext'))

from docs.dynamic_data import DynamicDataDirective

# Run django's initialization code so that all modules can be loaded
# but disable configuration of logging as it can break sphinx-build
os.environ["DJANGO_SETTINGS_MODULE"] = "debusine.project.settings"
settings.LOGGING_CONFIG = None
django.setup()

# -- Project information -----------------------------------------------------

project = 'Debusine'
copyright = f'2019-{dt.date.today().year}, The Debusine Developers'
author = 'The Debusine Developers'

# The short X.Y version
version = ''
# The full version, including alpha/beta/rc tags
release = ''


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'debusine_sphinx',
    'sphinx.ext.autodoc',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinx.ext.graphviz',
    'sphinx_copybutton',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# The name of the default domain.
primary_domain = 'debusine'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None

# Tweak copybutton so that it copies only commands and not prompts or outputs
copybutton_exclude = '.linenos, .gp'
copybutton_here_doc_delimiter = "END"
copybutton_line_continuation_character = "\\"
copybutton_prompt_text = r"\$ |# |root@debusine:~# "
copybutton_prompt_is_regexp = True
copybutton_only_copy_prompt_lines = True

# https://github.com/sphinx-doc/sphinx/issues/14223
suppress_warnings = ["ref.python"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'logo_only': True,
    'style_nav_header_background': '#757575',
}
html_logo = 'images/debusine-logo-with-white-name.png'
html_favicon = 'images/favicon.png'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
html_sidebars = {
    '*': [
        'globaltoc.html',
        'relations.html',
        'sourcelink.html',
        'searchbox.html',
    ],
}

html_baseurl = "https://freexian-team.pages.debian.net/debusine/"


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'Debusinedoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        'Debusine.tex',
        'Debusine Documentation',
        'Raphaël Hertzog and other Debusine contributors',
        'manual',
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'debusine', 'Debusine Documentation', [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        'Debusine',
        'Debusine Documentation',
        author,
        'Debusine',
        'One line description of project.',
        'Miscellaneous',
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {
    'python': ('https://docs.python.org/', None),
}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# If true, `todolist` produces output without file path and line.
todo_link_only = True

# -- Options for autodoc extension -------------------------------------------
autodoc_default_options = {
    'members': True,
    'member-order': 'alphabetical',
    'special-members': '__init__',
    'undoc-members': True,
    'exclude-members': '__weakref__',
    'show-inheritance': True,
}

# -- Options for linkcheck extension -----------------------------------------
linkcheck_ignore = [
    r'https://matrix\.to',
    r'http://localhost',
    r'https://diataxis\.fr',
]

linkcheck_anchors_ignore = [
    r"^!",
    # GitLab comments
    r"^note_[0-9]+$",
    # Hetzner documentation
    r"^tag/servers/create_server$",
]

linkcheck_exclude_documents = [
    r"^reference/release-history$",
]

linkcheck_request_headers = {
    # These sites seem to block non-browser User-Agents, but work if we look
    # at least a bit like a browser.
    "https://repost.aws/": {"User-Agent": "Mozilla/5.0"},
    "https://docs.hetzner.com/": {"User-Agent": "Mozilla/5.0"},
    "https://www.hetzner.com/": {"User-Agent": "Mozilla/5.0"},
    # Report accurate User-Agent to various *.debian.org sites, instead of
    # the one defined by Sphinx (sphinx/util/requests.py).
    "https://salsa.debian.org/": {
        "User-Agent": "Debusine-Linkcheck (+https://debusine.debian.net/; sphinx-linkcheck)"
    },
    "https://bugs.debian.org/": {
        "User-Agent": "Debusine-Linkcheck (+https://debusine.debian.net/; sphinx-linkcheck)"
    },
    "https://wiki.debian.org/": {
        "User-Agent": "Debusine-Linkcheck (+https://debusine.debian.net/; sphinx-linkcheck)"
    },
}

# -- Options for extlinks extension ------------------------------------------
extlinks = {
    "issue": (
        "https://salsa.debian.org/freexian-team/debusine/-/work_items/%s",
        "#%s",
    ),
    "mr": (
        "https://salsa.debian.org/freexian-team/debusine/-/merge_requests/%s",
        "!%s",
    ),
}


def setup(app):
    app.add_directive("dynamic_data", DynamicDataDirective)
    app.add_css_file("css/debusine.css")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: assets."""

from copy import copy
from logging import getLogger
from typing import Any, override

from django.core.exceptions import ObjectDoesNotExist
from rest_framework import status
from rest_framework.parsers import JSONParser
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import BaseSerializer

from debusine.artifacts.models import ArtifactCategory
from debusine.assets.models import AssetCategory
from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    Asset,
    AssetUsage,
    WorkRequest,
)
from debusine.db.models.assets import AssetQuerySet
from debusine.db.models.permissions import anonymous_access
from debusine.db.permissioncontext import PermissionContext
from debusine.server.exceptions import DebusineAPIException
from debusine.server.serializers import (
    AssetPermissionCheckRequest,
    AssetPermissionCheckResponse,
    AssetSerializer,
)
from debusine.server.views.base import (
    BaseAPIView,
    CanDisplayFilterBackend,
    GetOrCreateAPIView,
    ListAPIViewBase,
)
from debusine.server.views.rest import IsWorkRequestTokenAuthenticated
from debusine.tasks.models import WorkerType

log = getLogger(__name__)


class AssetView(GetOrCreateAPIView[Asset], ListAPIViewBase[Asset], BaseAPIView):
    """View used to get or create assets."""

    filter_backends = [CanDisplayFilterBackend]
    parser_classes = [JSONParser]
    serializer_class = AssetSerializer
    pagination_class = None

    @override
    def get_queryset(self) -> AssetQuerySet[Any]:
        """Get the query set for this view."""
        # queryset is filered to can_display
        self.enforce(anonymous_access)
        filter_kwargs: dict[str, Any] = {}
        if self.request.method == "GET":
            if asset_id := self.request.GET.get("asset", None):
                filter_kwargs["id"] = asset_id
            if work_request_id := self.request.GET.get("work_request", None):
                filter_kwargs["created_by_work_request_id"] = work_request_id
            if workspace := self.request.GET.get("workspace", None):
                filter_kwargs["workspace__name"] = workspace
            if not filter_kwargs:
                raise DebusineAPIException(
                    title=(
                        '"asset", "work_request", or "workspace" parameters '
                        'are mandatory'
                    )
                )
        return (
            Asset.objects.in_current_scope()
            .can_display(context.pc)
            .filter(**filter_kwargs)
            .order_by("id")
        )

    @override
    def get_existing_object(self, serializer: BaseSerializer[Asset]) -> Asset:
        """
        If there is an existing object matching Serializer, return it.

        If not, raise ObjectDoesNotExist.
        """
        workspace = serializer.validated_data["workspace"]
        self.set_current_workspace(workspace)
        self.enforce(workspace.can_display)
        self.enforce(workspace.can_create_artifacts)
        data = copy(serializer.validated_data)
        match data["category"]:
            case AssetCategory.SIGNING_KEY:
                obj = self.get_queryset().get(
                    category=data["category"],
                    data__purpose=data["data"].get("purpose", None),
                    data__fingerprint=data["data"].get("fingerprint", None),
                )
            case _:  # pragma: no cover
                raise NotImplementedError(
                    f"Unable to locate assets of category '{data['category']}'"
                )
        assert isinstance(obj, Asset)
        return obj

    @override
    def perform_create(self, serializer: BaseSerializer[Asset]) -> None:
        """Create an asset, and make it usable from its workspace."""
        asset = serializer.save()
        AssetUsage.objects.get_or_create(asset=asset, workspace=asset.workspace)


class AssetPermissionCheckView(BaseAPIView):
    """Check if a work request is allowed to use an asset."""

    permission_classes = [IsWorkRequestTokenAuthenticated]

    def get_object(self, asset_category: str, asset_slug: str) -> Asset:
        """Return the requested asset."""
        try:
            return Asset.objects.get_by_slug(
                category=asset_category,
                slug=asset_slug,
            )
        except Asset.DoesNotExist as exc:
            raise DebusineAPIException(
                title=str(exc), status_code=status.HTTP_404_NOT_FOUND
            )
        except ValueError as exc:
            raise DebusineAPIException(
                title=f"Malformed asset_slug: {exc}",
                status_code=status.HTTP_404_NOT_FOUND,
            )

    @classmethod
    def describe_artifact(cls, artifact: Artifact) -> dict[str, str]:
        """Return a dictionary description of artifact."""
        resource: dict[str, str] = {}
        match artifact.category:
            case ArtifactCategory.REPOSITORY_INDEX:
                # A debian:repository-index artifact must have exactly one
                # file.
                resource["path"] = artifact.fileinartifact_set.get().path
            case ArtifactCategory.SIGNING_INPUT:
                resource["package"] = artifact.data["binary_package_name"]
                try:
                    extracted_from = artifact.relations.get(
                        type=ArtifactRelation.Relations.RELATES_TO,
                        target__category=ArtifactCategory.BINARY_PACKAGE,
                    ).target
                except ArtifactRelation.DoesNotExist:
                    log.info(
                        (
                            "Unable to fully describe artifact %r, no related "
                            "debian:binary-package."
                        ),
                        artifact,
                    )
                except ArtifactRelation.MultipleObjectsReturned:
                    log.info(
                        (
                            "Unable to fully describe artifact %r, multiple "
                            "related debian:binary-package."
                        ),
                        artifact,
                    )
                else:
                    resource["source"] = extracted_from.data["srcpkg_name"]
                    deb_fields = extracted_from.data["deb_fields"]
                    resource["architecture"] = deb_fields["Architecture"]
                    if not resource["package"]:
                        resource["package"] = deb_fields["Package"]
                    resource["version"] = deb_fields["Version"]
            case ArtifactCategory.UPLOAD:
                match artifact.data["type"]:
                    case "dpkg":
                        changes_fields = artifact.data["changes_fields"]
                        for key in (
                            "Architecture",
                            "Source",
                            "Version",
                            "Distribution",
                        ):
                            resource[key.lower()] = changes_fields[key]
                    case _:
                        raise NotImplementedError(
                            f"Unable to describe upload artifact {artifact}"
                        )
            case _:
                raise NotImplementedError(
                    f"Unable to describe artifact {artifact}"
                )
        return resource

    def post(
        self,
        request: Request,
        asset_category: str,
        asset_slug: str,
        permission_name: str,
    ) -> Response:
        """Handle a POST request to check permissions."""
        # IsWorkerTokenAuthenticated has already passed
        work_request = context.pc.work_request
        assert work_request is not None
        assert work_request.worker is not None
        self.set_current_workspace(work_request.workspace)

        worker_type = work_request.worker.worker_type
        if worker_type != WorkerType.SIGNING:
            raise DebusineAPIException(
                title="Forbidden",
                detail=f"Endpoint is not accessible to {worker_type} workers",
                status_code=status.HTTP_403_FORBIDDEN,
            )

        check_request = AssetPermissionCheckRequest(data=request.data)
        check_request.is_valid(raise_exception=True)

        asset = self.get_object(
            asset_category=asset_category, asset_slug=asset_slug
        )

        try:
            artifact = Artifact.objects.get(
                id=check_request.validated_data["artifact_id"]
            )
            work_request = WorkRequest.objects.get(
                id=check_request.validated_data["work_request_id"]
            )
        except ObjectDoesNotExist as exc:
            raise DebusineAPIException(title=str(exc))

        self.check_object_permissions(request, work_request)
        self.enforce(work_request.can_display)
        self.check_object_permissions(request, artifact)
        self.enforce(artifact.can_display)

        # Evaluate permissions in the context of the work request, not just the
        # current user
        with context.local():
            context.reset()
            context.set_permission_context_scoped(
                work_request.workspace.scope,
                PermissionContext.for_work_request(work_request=work_request),
            )
            self.enforce(work_request.can_display)

        workspace = work_request.workspace
        if workspace.name != check_request.validated_data["workspace"]:
            raise DebusineAPIException(
                title="workspace does not match work request"
            )

        match permission_name:
            case "sign_with":
                predicate_name = "can_sign_with"
            case "sign_repository_with":
                predicate_name = "can_sign_repository_with"
            case _:
                raise DebusineAPIException(
                    title=f"Unknown permission '{permission_name}'"
                )

        try:
            asset_usage = asset.usage.get(workspace=workspace)
        except AssetUsage.DoesNotExist:
            has_permission = False
        else:
            has_permission = getattr(asset_usage, predicate_name)(
                PermissionContext.for_work_request(work_request=work_request)
            )

        response = AssetPermissionCheckResponse(
            data={
                "has_permission": has_permission,
                "username": work_request.created_by.username,
                "user_id": work_request.created_by.id,
                "resource": self.describe_artifact(artifact),
            }
        )
        response.is_valid(raise_exception=True)
        return Response(response.data, status=status.HTTP_200_OK)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Tests for generic templates or included templates.

The majority of template code is tested via test_views.py. Any other specific
code for templates is tested in this file.
"""

from typing import override

from django.contrib.auth import get_user_model
from django.core.handlers import exception
from django.template.loader import get_template
from django.urls import Resolver404

from debusine.db.models import WorkRequest
from debusine.test.django import TestCase


def _html_work_request_status(status: str) -> str:
    if status == "pending":
        text_bg = "text-bg-secondary"
        text = "Pending"
    elif status == "running":
        text_bg = "text-bg-info"
        text = "Running"
    elif status == "completed":
        text_bg = "text-bg-primary"
        text = "Completed"
    elif status == "aborted":
        text_bg = "text-bg-dark"
        text = "Aborted"
    elif status == "blocked":
        text_bg = "text-bg-dark"
        text = "Blocked"
    else:
        assert False

    return f'<span class="badge {text_bg}">{text}</span>'


class TemplateTests(TestCase):
    """Tests for the project level templates."""

    def test_404_render_exception(self) -> None:
        """404.html template renders the exception message."""
        template = get_template("404.html")
        exception_msg = "This is an exception message"
        context = {"exception": exception_msg}

        rendered_template = template.render(context)

        self.assertIn(exception_msg, rendered_template)

    def test_404_render_error(self) -> None:
        """404.html render error."""
        template = get_template("404.html")
        error = "This is an error message"
        context = {"error": error}

        self.assertIn(error, template.render(context))

    def test_404_no_error_no_useful_exception(self) -> None:
        """
        Template 404.html does not render exception.__class__.__name.__.

        In 404.html it's hardcoded as "Resolver404".
        """
        template = get_template("404.html")
        context = {"exception": Resolver404()}
        rendered_template = template.render(context)

        self.assertNotIn(str(exception.__class__.__name__), rendered_template)


class WorkRequestResultTests(TestCase):
    """Tests for _work_request-result.html."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.template = get_template("web/_work_request-result.html")

    def test_work_request_results(self) -> None:
        """Test rendering for possible results."""
        for result in ["success", "failure", "error", "skipped"]:
            with self.subTest(result=result):
                tree = self.assertHTMLValid(
                    self.template.render({"result": result})
                )
                self.assertTextContentEqual(tree, result.capitalize())

    def test_work_request_other(self) -> None:
        """Test for result is empty."""
        context = {"result": ""}
        rendered_template = self.template.render(context)
        self.assertHTMLEqual(rendered_template, "")


class WorkRequestResultSmallTests(TestCase):
    """Tests for _work_request-result-small.html."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.template = get_template("web/_work_request-result-small.html")

    def test_work_request_result_small(self) -> None:
        """Test rendering for different results."""
        wr = WorkRequest.Results

        def _render(css_class: str, title: str, text: str) -> str:
            """Return HTML."""
            return (
                f'<span title="{title}" class="badge {css_class}">{text}</span>'
            )

        result_to_html = {
            wr.SUCCESS: _render("text-bg-success", "Success", "S"),
            wr.FAILURE: _render("text-bg-warning", "Failure", "F"),
            wr.ERROR: _render("text-bg-danger", "Error", "E"),
            wr.SKIPPED: _render("text-bg-secondary", "Skipped", "s"),
            "unknown-result": "unknown-result",
        }

        for result, rendered in result_to_html.items():
            with self.subTest(result=repr(result)):
                rendered_template = self.template.render({"result": result})

                self.assertHTMLEqual(rendered_template, rendered)


class WorkRequestStatusTests(TestCase):
    """Tests for _work_request-status.html."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.template = get_template("web/_work_request-status.html")

    def test_work_request_status(self) -> None:
        """Test rendering for different status."""
        for status, _ in WorkRequest.Statuses.choices:
            with self.subTest(status=status):
                rendered_template = self.template.render({"status": status})

                self.assertHTMLEqual(
                    rendered_template, _html_work_request_status(status)
                )

    def test_work_request_status_aborted_with_user(self) -> None:
        """Aborted status shows aborting user when available."""
        User = get_user_model()
        user = User.objects.create_user(username="testuser")

        rendered_template = self.template.render(
            {
                "status": "aborted",
                "aborted_by": user,
            }
        )

        self.assertIn(
            'title="Aborted by testuser"',
            rendered_template,
        )

    def test_work_request_unexpected(self) -> None:
        """Test rendering function raise assertion."""
        with self.assertRaises(AssertionError):
            _html_work_request_status("")


class BadgeCountTests(TestCase):
    """Tests for _badge-count.html."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.template = get_template("web/_badge-count.html")

    def test_count_is_zero(self) -> None:
        """Test count is zero background is bg-secondary."""
        self.assertHTMLEqual(
            self.template.render(context={"count": 0, "title": "count of..."}),
            '<span class="badge text-bg-secondary"'
            ' title="count of...">0</span>',
        )

    def test_count_is_not_zero(self) -> None:
        """Test count is zero background is as passed."""
        self.assertHTMLEqual(
            self.template.render(
                context={
                    "count": 11,
                    "title": "count of...",
                    "bg_class": "primary",
                }
            ),
            (
                '<span class="badge text-bg-primary" '
                'title="count of...">11</span>'
            ),
        )


class DictToHtmlTests(TestCase):
    """Tests for _dict-to-html.html."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.template = get_template("web/_dict_to_ul.html")

    def test_recursive_dict_and_list(self) -> None:
        """Test recursive dict."""
        rendered = self.template.render(
            context={
                "dict": {
                    "keyA": "valueB",
                    "keyC": {
                        "keyD": "valueE",
                    },
                    "keyF": [1, 2],
                }
            }
        )
        self.assertHTMLEqual(
            rendered,
            """
            <ul>
                <li><b>keyA:</b> valueB</li>
                <li><b>keyC:</b>
                    <ul>
                        <li><b>keyD:</b> valueE</li>
                    </ul>
                </li>
                <li><b>keyF:</b>
                    <ul>
                        <li>1</li>
                        <li>2</li>
                    </ul>
                </li>
            </ul>
            """,
        )
